import collections

__all__ = ['SideEffect']


class SideEffect(object):
    '''
    Basic class to express side-effect actions.

    SideEffect in yawf is a side-effect action that can be performed
    after certain changes in workflow.

    User can treat effects as event handlers and put that handlers on
    any transition. To specify a transition, one can specify a list of
    states before transition, a list of states after transition and a list of
    message ids. Everything above is optional, so if you register a default
    side-effect without specify any of this variables, such side-effect will
    be called after *each* transition.

    User can put any number of side-effects on each and every transition. In
    case of multiple effects matching specific transition each side-effect
    will be executed one by one.

    Yawf executes side-effect *after* state transition, when yawf obtains new
    state (and possibly modified object).

    Yawf passes keyword arguments to side-effect:
      * `old_obj`: the object *before* the state transition (before locking);
      * `obj`: the object *after* the state transition;
      * `sender`: message sender;
      * `params`: message arguments;
      * `message_spec`: message spec class (because we may want to distinguish
            between different messages in effects);
      * `extra_context`: extra context thas was passed to
            :py:func:`yawf.dispatch.dispatch`;
      * `handler_result`: result of the state transition routine.
    '''

    message_id = None
    message_group = None
    states_from = None
    states_to = None
    is_transactional = False

    def __init__(self, message_id=None,
            states_from=None, states_to=None,
            message_group=None):

        if message_id is not None:
            self.message_id = message_id
        if message_group is not None:
            self.message_group = message_group
        if states_to is not None:
            self.states_to = states_to
        if states_from is not None:
            self.states_from = states_from

        message_group = self.message_group

        if message_group is not None:
            if isinstance(message_group, basestring):
                self.message_group = [message_group]
            else:
                assert isinstance(message_group, collections.Iterable)

        super(SideEffect, self).__init__()

    def perform(self, **kwargs):
        return kwargs

    def set_performer(self, performer):
        self.perform = lambda **kwargs: performer(**kwargs)

    def __call__(self, **kwargs):
        return self.perform(**kwargs)

    @property
    def name(self):
        return self.__class__.__name__


# for backward-compatibility
SideEffectAction = SideEffect



from django.conf import settings

DEFAULT_CONFIG = {
    'WORKFLOW_TYPE_ATTR': 'workflow_type',
    'INITIAL_STATE': 'init',
    'STATE_TYPE_CONSTRAINT': basestring,
    'DEFAULT_START_MESSAGE': 'start_workflow',
    'SOFT_DELETE_ATTR': None,
    'REVISION_ATTR': 'revision',
    'MESSAGE_LOG_ENABLED': False,
    'TRANSACTIONAL_SIDE_EFFECT': True,
    'USE_SELECT_FOR_UPDATE': True,
    'REVISION_BACKEND':
        'yawf.revision.backends.reversion.ReversionRevisionManager',
}

CONFIG = DEFAULT_CONFIG.copy()

custom_config = getattr(settings, 'YAWF_CONFIG', None)

if custom_config:
    CONFIG.update(custom_config)

locals().update(CONFIG)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding field 'MessageLog.parent_id'
        db.add_column('message_log_messagelog', 'parent_id', self.gf('django.db.models.fields.related.ForeignKey')(to=orm['message_log.MessageLog'], null=True, blank=True), keep_default=False)


    def backwards(self, orm):
        
        # Deleting field 'MessageLog.parent_id'
        db.delete_column('message_log_messagelog', 'parent_id_id')


    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'message_log.messagelog': {
            'Meta': {'ordering': "('created_at',)", 'object_name': 'MessageLog'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'message_logs_instance'", 'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initiator_content_type': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'message_logs_initiator'", 'null': 'True', 'to': "orm['contenttypes.ContentType']"}),
            'initiator_object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'message': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'message_params': ('django.db.models.fields.TextField', [], {}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'parent_id': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['message_log.MessageLog']", 'null': 'True', 'blank': 'True'}),
            'revision_after': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'pre_message'", 'null': 'True', 'to': "orm['revision.Revision']"}),
            'revision_before': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'post_message'", 'null': 'True', 'to': "orm['revision.Revision']"}),
            'workflow_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '64', 'db_index': 'True'})
        },
        'revision.revision': {
            'Meta': {'unique_together': "[('object_id', 'content_type', 'revision')]", 'object_name': 'Revision'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'message_log_record': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'affected_revisions'", 'null': 'True', 'to': "orm['message_log.MessageLog']"}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'revision': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'serialized_fields': ('django.db.models.fields.TextField', [], {})
        }
    }

    complete_apps = ['message_log']



from itertools import ifilter
from collections import defaultdict, Iterable
from operator import attrgetter
from functools import wraps, partial
import types

from yawf.config import STATE_TYPE_CONSTRAINT
from yawf import get_workflow_by_instance


def chained_apply(callables_iterable):
    def chained_wrapper(*args, **kwargs):
        for fun in callables_iterable:
            fun(*args, **kwargs)
    return chained_wrapper


def make_common_updater(kwargs, field_names=None, post_hook=None):

    # to ensure that we will update, not insert new
    kwargs.pop('id', None)
    kwargs.pop('pk', None)

    if field_names:
        def common_updater(obj):
            for field_name in field_names:
                value = kwargs.get(field_name, '__missing')
                if value != '__missing':
                    setattr(obj, field_name, value)

            obj.save()

            if callable(post_hook):
                return post_hook(obj)
    else:
        def common_updater(obj):

            for field_name in (f.name for f in obj._meta.fields):
                value = kwargs.get(field_name, '__missing')
                if value != '__missing':
                    setattr(obj, field_name, value)

            for field_name in (f.name for f in obj._meta.many_to_many):
                value = kwargs.get(field_name)
                if value is not None:
                    setattr(obj, field_name, value)

            obj.save()

            if callable(post_hook):
                return post_hook(obj)

    return common_updater


def common_cancel(obj, cancel_state='canceled', soft_delete_attr=None):
    obj.state = cancel_state
    if soft_delete_attr:
        setattr(obj, soft_delete_attr, True)
    obj.save()


def common_start(obj, state, soft_delete_attr=None):
    obj.state = state
    if soft_delete_attr:
        setattr(obj, soft_delete_attr, False)
    obj.save()


def make_common_cancel(cancel_state='canceled', soft_delete_attr=None):

    return partial(common_cancel,
        cancel_state=cancel_state,
        soft_delete_attr=soft_delete_attr)


def make_common_start(state, soft_delete_attr=None):

    return partial(common_start,
        state=state, soft_delete_attr=soft_delete_attr)


def optionally_edit(handler):

    @wraps(handler)
    def wrapper(obj, sender, edit_fields=None, **kwargs):

        handler_result = handler(obj, sender, **kwargs)
        workflow = get_workflow_by_instance(obj)

        if not edit_fields:
            return handler_result
        else:
            if isinstance(handler_result, STATE_TYPE_CONSTRAINT):
                edit_fields['state'] = handler_result
                return workflow.make_updater(edit_fields)
            else:
                assert callable(handler_result)
                updater = workflow.make_updater(edit_fields)
                return chained_apply((updater, handler_result))

    return wrapper


def select_for_update(queryset):
    return queryset.select_for_update()


def model_diff(instance1, instance2, full=False):
    diff = []

    if full:
        fields = instance1._meta.fields
    else:
        # Check only editable fields in model
        fields = ifilter(attrgetter('editable'), instance1._meta.fields)

    for field in fields:
        field_name = field.name
        value1 = getattr(instance1, field_name)
        value2 = getattr(instance2, field_name)
        if value1 != value2:
            diff.append({'field_name': field_name,
                    'field_verbose_name': field.verbose_name,
                    'old': value1, 'new': value2})

    return diff


def model_diff_fields(instance1, instance2):
    # Check only editable fields in model
    for field in ifilter(attrgetter('editable'), instance1._meta.fields):

        field_name = field.name
        value1 = getattr(instance1, field_name)
        value2 = getattr(instance2, field_name)
        if value1 != value2:
            yield field_name


def memoizible_property(getter):

    key = '_cache_prop_' + getter.__name__

    def getter_wrapper(self, *args, **kwargs):

        if not hasattr(self, key):
            result = getter(self, *args, **kwargs)
            if isinstance(result, types.GeneratorType):
                result = list(result)
            setattr(self, key, result)
            return result
        return getattr(self, key)

    return property(getter_wrapper)


def maybe_list(a):

    if a is not None:
        if isinstance(a, basestring) or not isinstance(a, Iterable):
            return [a]
        else:
            return list(a)
    else:
        return []


def metadefaultdict(fabric):
    return lambda: defaultdict(fabric)


def filter_handler_result(result, cls_to_filter):
    if isinstance(result, Iterable):
        return [
            r for r in result
            if isinstance(r, cls_to_filter)
        ]
    else:
        return []


def filter_side_effect_results(results, cls_to_filter):
    filtered = []
    for result in results:
        if isinstance(result, Iterable):
            filtered.extend(
                r for r in result
                if isinstance(r, cls_to_filter)
            )
    return filtered



# Create your views here.
from django import http

from yawf.views import YawfMessageView, HandlerViewMixin

from yawf_sample.simple.models import Window


class ResizeView(YawfMessageView):

    workflow_type = 'simple'
    model = Window
    message_id = 'edit__resize'

    def get_success_url(self):
        return '/'


class ToMaximizedView(HandlerViewMixin, YawfMessageView):

    workflow_type = 'simple'
    states_from = 'normal', 'minimized'

    def get_message_id(self):
        return 'maximize'

    def perform(self, obj, sender):
        return 'maximized'

    def wrap_yawf_result(self, obj, hr, er):
        return http.HttpResponse('OK')


window_resize = ResizeView.as_view()
window_maximize = ToMaximizedView.as_view()



from django.core.exceptions import ImproperlyConfigured
from django.utils.importlib import import_module

from yawf.config import REVISION_BACKEND


def get_class_from_dotted_path(import_path):

    """
    Imports the message storage class described by import_path, where
    import_path is the full Python path to the class.
    """
    try:
        dot = import_path.rindex('.')
    except ValueError:
        raise ImproperlyConfigured("%s isn't a Python path." % import_path)
    module, classname = import_path[:dot], import_path[dot + 1:]
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured('Error importing module %s: "%s"' %
                                   (module, e))
    try:
        return getattr(mod, classname)
    except AttributeError:
        raise ImproperlyConfigured('Module "%s" does not define a "%s" '
                                   'class.' % (module, classname))


default_revision_manager = get_class_from_dotted_path(REVISION_BACKEND)
from .models import RevisionModelMixin



# -*- coding: utf-8 -*-
from yawf import get_workflow_by_instance


def get_allowed(sender, obj):
    workflow = get_workflow_by_instance(obj)

    obj_state = getattr(obj, workflow.state_attr_name)

    check_result = dict(
        (c, c(obj, sender))
        for c in workflow.get_checkers_by_state(obj_state))

    messages = []
    for checker, message in workflow.get_available_messages(obj_state):
        if checker(obj, sender, cache=check_result):
            spec = workflow.get_message_spec(message)
            messages.append({'id': spec.id, 'title': unicode(spec), 'rank': spec.rank})

    resources = []
    for resource in workflow.get_available_resources(obj_state):
        if resource.permission_checker(obj, sender, cache=check_result):
            resources.append(
                {
                    'id': resource.id,
                    'description': resource.description,
                    'slug': resource.slug,
                })

    return {'allowed_messages': messages,
            'allowed_resources': resources}



from django.dispatch import Signal

message_handled = Signal(
    providing_args=[
        'workflow',
        'message',
        'instance',
        'new_instance',
        'transition_result',
        'side_effect_result',
        'log_record',
    ])

transition_handled = Signal(
    providing_args=[
        'workflow',
        'message',
        'instance',
        'new_instance',
        'transition_result'
    ])



from django.utils.encoding import force_unicode

from yawf import get_workflow_by_instance


class WorkflowAwareModelBase(object):

    _workflow = None

    @property
    def workflow(self):
        if not self._workflow:
            self._workflow = get_workflow_by_instance(self)
        return self._workflow

    def workflow_type_display(self):
        return self.workflow.verbose_name

    def state_display(self):
        return force_unicode(self.workflow.verbose_state_names.get(self.state)) or self.state

    def get_clarified_instance(self):
        workflow = self.workflow

        if workflow.model_class is self.__class__:
            return self
        else:
            return workflow.model_class.objects.get(id=self.id)



from __future__ import absolute_import
import reversion

from . import RevisionManager

# revision_merger is a little hack for reversion to "merge" optional
# revision information to records
class ReversionMerger(object):

    def __init__(self, attrname):
        self.attrname = attrname
        super(ReversionMerger, self).__init__()

    def create(self, revision, obj):
        setattr(obj, self.attrname, revision)
        obj.save()

    def db_manager(self, db):
        return self

    @property
    def _default_manager(self):
        return self


class ReversionRevisionManager(RevisionManager):

    def __enter__(self):
        self._reversion_manager = reversion.create_revision()
        self._reversion_manager.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._reversion_manager.__exit__(exc_type, exc_value, traceback)

    def bind_revision(self, obj, attrname='revision'):
        reversion.add_meta(ReversionMerger(attrname), obj=obj)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from yawf.config import INITIAL_STATE
from yawf.base_model import WorkflowAwareModelBase


class WorkflowAwareModel(WorkflowAwareModelBase, models.Model):

    class Meta:
        abstract = True

    state = models.CharField(default=INITIAL_STATE,
            max_length=32, db_index=True, editable=False,
            verbose_name=_('state'))



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Adding field 'MessageLog.revision_content_type'
        db.add_column('message_log_messagelog', 'revision_content_type',
                      self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='message_logs_revision', null=True, to=orm['contenttypes.ContentType']),
                      keep_default=False)

        # Adding field 'MessageLog.revision_id'
        db.alter_column('message_log_messagelog', 'revision_id',
                      self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True, null=True, blank=True),
                      explicit_name=True)


    def backwards(self, orm):
        # Adding field 'MessageLog.revision'
        db.alter_column('message_log_messagelog', 'revision_id',
                      self.gf('django.db.models.fields.related.ForeignKey')(related_name='message_log', null=True, to=orm['reversion.Revision'], blank=True),
                      explicit_name=True)

        # Deleting field 'MessageLog.revision_content_type'
        db.delete_column('message_log_messagelog', 'revision_content_type_id')


    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'message_log.messagelog': {
            'Meta': {'ordering': "('created_at',)", 'object_name': 'MessageLog'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'message_logs_instance'", 'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'group_uuid': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '36', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initiator_content_type': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'message_logs_initiator'", 'null': 'True', 'to': "orm['contenttypes.ContentType']"}),
            'initiator_object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'message': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'message_params': ('django.db.models.fields.TextField', [], {}),
            'message_params_dehydrated': ('django.db.models.fields.TextField', [], {'default': "''"}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'parent_uuid': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '36', 'null': 'True', 'blank': 'True'}),
            'revision_content_type': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'message_logs_revision'", 'null': 'True', 'to': "orm['contenttypes.ContentType']"}),
            'revision_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'transition_result': ('django.db.models.fields.TextField', [], {'default': "''"}),
            'uuid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '36', 'db_index': 'True'}),
            'workflow_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '64', 'db_index': 'True'})
        }
    }

    complete_apps = ['message_log']



from django.http import Http404, HttpResponse

from yawf import get_workflow
from yawf.graph import build_effects_graph, build_handlers_graph


def handlers_graph(request, workflow_id):
    return _graph_view(request, workflow_id, build_handlers_graph)


def effects_graph(request, workflow_id):
    return _graph_view(request, workflow_id, build_effects_graph)


def _graph_view(request, workflow_id, build_func):
    w = get_workflow(workflow_id)
    if w is None:
        raise Http404

    response = HttpResponse(mimetype='image/png')

    graph = build_func(w)
    graph.layout('dot')
    graph.draw(response, format='png')

    return response



from yawf.exceptions import ResourcePermissionDeniedError


class WorkflowResource(object):

    def __init__(self, handler, resource_id, permission_checker,
                                                description=None,
                                                slug=None):
        super(WorkflowResource, self).__init__()
        self._handler = handler
        self.id = resource_id
        self.permission_checker = permission_checker
        self.description = description
        self.slug = slug or self.id

    def __call__(self, request, obj, sender):

        if self.permission_checker(obj, sender):
            return self._handler(request, obj, sender)
        else:
            raise ResourcePermissionDeniedError(self.id,
                    obj, sender)



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding model 'MessageLog'
        db.create_table('message_log_messagelog', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('created_at', self.gf('django.db.models.fields.DateTimeField')(auto_now_add=True, db_index=True, blank=True)),
            ('content_type', self.gf('django.db.models.fields.related.ForeignKey')(related_name='message_logs_instance', to=orm['contenttypes.ContentType'])),
            ('object_id', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True)),
            ('initiator_content_type', self.gf('django.db.models.fields.related.ForeignKey')(blank=True, related_name='message_logs_initiator', null=True, to=orm['contenttypes.ContentType'])),
            ('initiator_object_id', self.gf('django.db.models.fields.PositiveIntegerField')(db_index=True, null=True, blank=True)),
            ('message', self.gf('django.db.models.fields.CharField')(max_length=32)),
            ('message_params', self.gf('django.db.models.fields.TextField')()),
            ('revision_before', self.gf('django.db.models.fields.related.ForeignKey')(related_name='post_message', null=True, to=orm['revision.Revision'])),
            ('revision_after', self.gf('django.db.models.fields.related.ForeignKey')(related_name='pre_message', null=True, to=orm['revision.Revision'])),
        ))
        db.send_create_signal('message_log', ['MessageLog'])


    def backwards(self, orm):
        
        # Deleting model 'MessageLog'
        db.delete_table('message_log_messagelog')


    models = {
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'message_log.messagelog': {
            'Meta': {'ordering': "('created_at',)", 'object_name': 'MessageLog'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'message_logs_instance'", 'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initiator_content_type': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'message_logs_initiator'", 'null': 'True', 'to': "orm['contenttypes.ContentType']"}),
            'initiator_object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'message': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'message_params': ('django.db.models.fields.TextField', [], {}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'revision_after': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'pre_message'", 'null': 'True', 'to': "orm['revision.Revision']"}),
            'revision_before': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'post_message'", 'null': 'True', 'to': "orm['revision.Revision']"})
        },
        'revision.revision': {
            'Meta': {'unique_together': "[('object_id', 'content_type', 'revision')]", 'object_name': 'Revision'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'message_log_record': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'affected_revisions'", 'null': 'True', 'to': "orm['message_log.MessageLog']"}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'revision': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'serialized_fields': ('django.db.models.fields.TextField', [], {})
        }
    }

    complete_apps = ['message_log']



# -*- coding: utf-8 -*-
import logging
import copy
from itertools import ifilter

from django.utils.encoding import smart_unicode

from yawf.message_log.models import log_message

from yawf.config import STATE_TYPE_CONSTRAINT,\
         TRANSACTIONAL_SIDE_EFFECT, USE_SELECT_FOR_UPDATE, MESSAGE_LOG_ENABLED
from yawf.exceptions import IllegalStateError,\
         WrongHandlerResultError, PermissionDeniedError,\
         MessageIgnored
from yawf.signals import message_handled
from yawf import get_workflow_by_instance
from yawf.messages import Message
from yawf.state_transition import transition, transactional_transition
from yawf.revision import default_revision_manager

logger = logging.getLogger(__name__)


class Dispatcher(object):
    '''
    Class-style dispatching.

    One can use Dispatcher instead of dispatch if it seems more natural.

    Example:

    >>> d = Dispatcher(obj, user, extra_context={'request': request})
    >>> new_obj, handler_result, effect_result = d.edit(foo='bar')

    So object-recipient and sender are passed to the dispatcher once and
    message passing is done by calling methods on dispatcher object.
    '''
    def __init__(self, obj, sender, **options):
        self.obj = obj
        self.sender = sender
        self.options = options
        super(Dispatcher, self).__init__()

    def __getattribute__(self, name):
        message_id = name
        return lambda **raw_params:\
            dispatch(self.obj, self.sender, message_id,
                    raw_params,
                    **self.options)


def dispatch(obj, sender, message_id,
        raw_params=None,
        extra_context=None):
    '''
    High-level shortcut function to send a message to workflow-enabled object.

    See :py:func:`dispatch_message` for detailed information.
    '''
    message = Message(sender, message_id, raw_params)
    return dispatch_message(
        obj, message, extra_context=extra_context)


def dispatch_no_clean(obj, sender, message_id, params=None, extra_context=None):
    message = Message(sender, message_id, clean_params=params)
    return dispatch_message(
        obj, message, extra_context=extra_context)


def dispatch_message(obj, message, extra_context=None,
        transactional_side_effect=TRANSACTIONAL_SIDE_EFFECT,
        need_lock_object=USE_SELECT_FOR_UPDATE,
        defer_side_effect=False,
        revision_manager=None):
    '''
    Gets an object and message and performs all actions specified by
    object's workflow.

    :param obj:
        Object that receives message. Must be an instance of
        :py:class:`yawf.base_model.WorkflowAwareModelBase`
    :param message:
        :py:class:`yawf.messages.Message` instance that incapsulates message sender and parameters

    :return:
        Tuple of three values:
         * new object instance (after state transition)
         * transition result (returned by handler object)
         * side effects results
    '''
    logger.debug(u"Backend got message from %s to %s: %s %s",
            smart_unicode(message.sender), smart_unicode(obj),
            message.id, smart_unicode(message.raw_params))

    # can raise WorkflowNotLoadedError
    workflow = get_workflow_by_instance(obj)

    # validate data and filter out trash
    message.clean(workflow, obj)

    # dehydrate message params for serializing
    message.dehydrate_params(workflow, obj)

    # find a transition handler, can raise handler-related errors
    handler = get_handler(workflow, message, obj)

    # fetch a transition, can raise app-specific handler errors
    handler_result = apply(handler, (obj, message.sender), message.params)

    # if handler returns None - do nothing
    if handler_result is None:
        raise MessageIgnored(message)

    # if handler returns type appropriate for state (string) - change state
    if isinstance(handler_result, STATE_TYPE_CONSTRAINT):
        if workflow.is_valid_state(handler_result):
            def state_transition(obj):
                setattr(obj, workflow.state_attr_name, handler_result)
                obj.save()
                return obj
        else:
            raise IllegalStateError(handler_result)

    # if handler returns callable, perform it as single transaction
    elif callable(handler_result):
        state_transition = handler_result
    else:
        raise WrongHandlerResultError(handler_result)

    if defer_side_effect:
        transition_ = transactional_transition
        transactional_side_effect = False
    else:
        transition_ = transition

    if revision_manager is None:
        revision_manager = default_revision_manager

    with revision_manager() as m:
        new_obj, transition_result, side_effect_result =\
            transition_(
                workflow, obj, message, state_transition,
                extra_context=extra_context,
                transactional_side_effect=transactional_side_effect,
                need_lock_object=need_lock_object)

        if MESSAGE_LOG_ENABLED:
            log_record = log_message(
                sender=workflow.id,
                workflow=workflow,
                message=message,
                instance=obj,
                new_instance=new_obj,
                transition_result=transition_result)

            m.bind_revision(log_record)
        else:
            log_record = None

    # TODO: send_robust + logging?
    message_handled.send(
            sender=workflow.id,
            workflow=workflow,
            message=message,
            instance=obj,
            new_instance=new_obj,
            transition_result=transition_result,
            side_effect_result=side_effect_result,
            log_record=log_record)

    return new_obj, transition_result, side_effect_result


def get_handler(workflow, message, obj):

    current_state = getattr(obj, workflow.state_attr_name)
    handlers = workflow.library.get_handlers(current_state, message.id)

    # check permission for a sender
    permitted_handlers = ifilter(
        lambda handler: handler.permission_checker(obj, message.sender),
        handlers)

    try:
        # NOTE: we will get permitted handler that was registered *first*,
        # if there are more than one handler for that state,message
        handler = permitted_handlers.next()
    except StopIteration:
        raise PermissionDeniedError(obj, message)
    else:
        if handler.copy_before_call:
            handler = copy.deepcopy(handler)
        return handler



from django.test import TestCase

from yawf.permissions import C, allow_to_all, restrict_to_all

__all__ = ('PermissionsTestCase',)


class PermissionsTestCase(TestCase):

    @staticmethod
    def call_count(func):
        func.call_count = 0
        def wrapper(*args):
            func.call_count += 1
            return func(*args)
        wrapper.func = func
        return wrapper

    def setUp(self):
        self.sender_is_even = self.call_count(lambda obj, sender: sender & 1 == 0)
        self.obj_is_even = self.call_count(lambda obj, sender: obj & 1 == 0)

        self.complex_checker = (
            C(self.obj_is_even, self.sender_is_even) |
                allow_to_all & ~C(self.sender_is_even) & allow_to_all)

    def test_custom_sender(self):
        self.assertFalse(self.sender_is_even(None, 3))
        self.assertTrue(self.sender_is_even(None, 4))
        self.assertFalse(self.obj_is_even(3, None))
        self.assertTrue(self.obj_is_even(4, None))

    def test_inversion(self):
        sender_is_odd = ~C(self.sender_is_even)

        self.assertTrue(sender_is_odd(None, 3))
        self.assertFalse(sender_is_odd(None, 4))

        sender_is_even = ~C(sender_is_odd)
        self.assertFalse(sender_is_even(None, 3))
        self.assertTrue(sender_is_even(None, 4))

    def test_inversion_invariance(self):
        sender_is_even = C(self.sender_is_even)
        sender_is_odd = ~sender_is_even
        sender_is_even_again = ~sender_is_odd
        self.assertTrue(sender_is_even is sender_is_even_again)

    def test_checker_expressions(self):
        complex_checker = self.complex_checker
        self.assertTrue(complex_checker(2, 2))
        self.assertTrue(complex_checker(0, 1))
        self.assertFalse(complex_checker(1, 2))

        new_allow_to_all = allow_to_all | complex_checker
        self.assertTrue(new_allow_to_all(2, 2))
        self.assertTrue(new_allow_to_all(0, 1))
        self.assertTrue(new_allow_to_all(1, 2))

        new_restrict_to_all = complex_checker & restrict_to_all
        self.assertFalse(new_restrict_to_all(2, 2))
        self.assertFalse(new_restrict_to_all(0, 1))
        self.assertFalse(new_restrict_to_all(1, 2))

    def test_atomics(self):
        complex_checker = self.complex_checker
        self.assertEqual(len(set(complex_checker.get_atomical_checkers())), 3)

    def test_cache(self):
        '''
        Call complex checker three times and check times that checker
        functions actually been called
        '''
        complex_checker = self.complex_checker
        self.assertEqual(self.sender_is_even.func.call_count, 0)
        self.assertEqual(self.obj_is_even.func.call_count, 0)
        complex_checker(0, 1)
        self.assertEqual(self.sender_is_even.func.call_count, 1)
        self.assertEqual(self.obj_is_even.func.call_count, 1)
        complex_checker(2, 2)
        self.assertEqual(self.sender_is_even.func.call_count, 2)
        self.assertEqual(self.obj_is_even.func.call_count, 2)
        complex_checker(1, 2)
        self.assertEqual(self.sender_is_even.func.call_count, 3)
        self.assertEqual(self.obj_is_even.func.call_count, 3)

    def test_fill_cache(self):
        cache = self.complex_checker.fill_cache(2, 2)
        allower = allow_to_all.get_atomical_checkers().next()
        self.assertEqual(len(cache), 3)
        self.assertTrue(self.sender_is_even in cache)
        self.assertTrue(self.obj_is_even in cache)
        self.assertTrue(allower in cache)

    def test_cyclic(self):
        or_checker = C(self.sender_is_even) | C(self.obj_is_even)
        atom_checkers = list(or_checker.get_atomical_checkers())
        or_checker |= or_checker
        self.assertListEqual(
            atom_checkers,
            list(or_checker.get_atomical_checkers()))

        and_checker = C(self.sender_is_even, self.obj_is_even)
        and_checker &= and_checker
        atom_checkers = list(and_checker.get_atomical_checkers())
        self.assertListEqual(
            atom_checkers,
            list(and_checker.get_atomical_checkers()))



# -*- coding: utf-8 -*-
from operator import attrgetter

from yawf.exceptions import (
        WorkflowNotLoadedError,
        WorkflowAlreadyRegisteredError)
from yawf.config import WORKFLOW_TYPE_ATTR

_registry = {}
_by_rank = []


# function to use by workflow objects to register themselves
def register_workflow(workflow):
    pass # this is done in workflow constructor now


def _register_workflow(workflow):

    if workflow.id in _registry:
        raise WorkflowAlreadyRegisteredError(workflow.id, workflow)

    _registry[workflow.id] = workflow
    _by_rank.append(workflow)
    _by_rank.sort(key=attrgetter('rank'))


# external api
get_registered_workflows = lambda: _by_rank
get_workflow = lambda workflow_id: _registry.get(workflow_id)

class WorkflowChoices(object):

    def __iter__(self):
        return get_workflow_name_map().iteritems()

choices = WorkflowChoices()


def get_workflow_name_map():
    return dict((workflow.id, workflow.verbose_name) for workflow
            in _registry.itervalues())


def get_workflow_display_name(id):
    return get_workflow(id).verbose_name


def get_workflow_by_instance(obj):
    workflow_type = getattr(obj, WORKFLOW_TYPE_ATTR)
    workflow = get_workflow(workflow_type)
    if workflow is None:
        raise WorkflowNotLoadedError(workflow_type)
    return workflow


def autodiscover():
    # borrowed from admin.autodiscover
    from django.conf import settings
    from django.utils.importlib import import_module

    for app in settings.INSTALLED_APPS:
        mod = import_module(app)
        # Attempt to import the app's workflow module.
        _try_to_import(mod, app, 'workflow')
        _try_to_import(mod, app, 'workflows')


def _try_to_import(mod, app, module_name):

    global _registry

    import copy
    from django.utils.importlib import import_module
    from django.utils.module_loading import module_has_submodule

    try:
        before_import_registry = copy.copy(_registry)
        import_module('.'.join([app, module_name]))
    except:
        # Reset the model registry to the state before the last import as
        # this import will have to reoccur on the next request and this
        # could raise NotRegistered and AlreadyRegistered exceptions
        # (see #8245).
        _registry = before_import_registry

        # Decide whether to bubble up this error. If the app just
        # doesn't have an admin module, we can ignore the error
        # attempting to import it, otherwise we want it to bubble up.
        if module_has_submodule(mod, module_name):
            raise



from . import Message

class Submessage(object):

    need_lock_object = True

    def __init__(self, obj, message_id, sender,
            params=None, need_lock_object=True, raw_params=None):
        self.obj = obj
        self.sender = sender
        self.message_id = message_id
        self.params = params
        self.raw_params = raw_params
        self.need_lock_object = need_lock_object
        super(Submessage, self).__init__()

    def as_message(self, parent):

        return Message(self.sender, self.message_id,
            raw_params=self.raw_params,
            clean_params=self.params,
            parent_message_id=parent.unique_id,
            message_group=parent.message_group,
        )

    def dispatch(self, parent_obj, parent_message):
        from yawf.dispatch import dispatch_message
        message = self.as_message(parent_message)
        return dispatch_message(
            self.obj,
            message=message,
            defer_side_effect=True,
            need_lock_object=self.need_lock_object)


class RecursiveSubmessage(Submessage):

    def __init__(self, message_id, sender, params=None, raw_params=None):
        super(RecursiveSubmessage, self).__init__(
            obj=None,
            sender=sender, message_id=message_id,
            params=params, raw_params=raw_params)

    def dispatch(self, parent_obj, parent_message):
        from yawf.dispatch import dispatch_message
        message = self.as_message(parent_message)
        return dispatch_message(
            parent_obj,
            message=message,
            defer_side_effect=True,
            need_lock_object=False)



from django.db import models
import reversion
from yawf.revision import RevisionModelMixin


class WINDOW_OPEN_STATUS:

    MINIMIZED = 'minimized'
    MAXIMIZED = 'maximized'
    NORMAL = 'normal'

    types = (MINIMIZED, MAXIMIZED, NORMAL)
    choices = zip(types, types)


class Window(RevisionModelMixin, models.Model):

    title = models.CharField(max_length=255)
    width = models.IntegerField()
    height = models.IntegerField()
    parent = models.ForeignKey('self', related_name='children',
            null=True, blank=True)

    workflow_type = 'simple'

    open_status = models.CharField(
        max_length=32,
        choices=WINDOW_OPEN_STATUS.choices,
        default='init',
        editable=False)

reversion.register(Window)



from django.template.response import TemplateResponse
from django.http import Http404, HttpResponseRedirect
from django.views.generic.edit import ProcessFormView
from django.views.generic.detail import SingleObjectMixin

from yawf import get_workflow
from yawf import dispatch
from yawf.exceptions import MessageValidationError


class MessageViewMixin(object):

    message_id = None

    def get_message_id(self):

        dynamic_message_id = self.kwargs.get('message_id')
        return self.message_id\
            if dynamic_message_id is None else dynamic_message_id

    def get_sender(self, *args, **kwargs):

        return self.request.user


class YawfMessageView(MessageViewMixin, SingleObjectMixin, ProcessFormView):

    @property
    def model(self):
        workflow = get_workflow(self.workflow_type)
        if hasattr(workflow, 'model_class'):
            return workflow.model_class

    def get_yawf_object(self):
        return self.object

    def post(self, request, *args, **kwargs):

        view_obj = self.get_object()
        self.object = view_obj

        obj = self.get_yawf_object()

        if hasattr(obj, 'get_clarified_instance'):
            obj = obj.get_clarified_instance()

        self.yawf_object = obj

        msg_id = self.get_message_id()
        sender = self.get_sender()

        try:
            obj, handler_result, effect_result = dispatch.dispatch(obj, sender,
                    msg_id, self.request.POST)
        except BaseException as e:
            return self.process_exception(obj, sender, msg_id, e)
        else:
            return self.wrap_yawf_result(obj, handler_result, effect_result)

    def wrap_yawf_result(self, obj, handler_result, effect_result):
        return HttpResponseRedirect(self.get_success_url())

    def process_exception(self, obj, sender, msg_id, exc):
        if isinstance(exc, MessageValidationError):
            return self.form_invalid(exc.validator)
        else:
            raise


class HandlerViewMixin(MessageViewMixin):

    states_from = None
    permission_checker = None
    workflow_type = None

    @classmethod
    def as_view(cls, **initkwargs):
        view = super(HandlerViewMixin, cls).as_view(**initkwargs)

        instance = cls(**initkwargs)
        workflow = get_workflow(cls.workflow_type)
        workflow.library.handler(
            message_id=instance.get_message_id(),
            states_from=instance.states_from,
            permission_checker=instance.permission_checker)(instance.perform)

        return view

    def perform(self, obj, sender, **kwargs):
        return lambda obj: self.transition(obj, sender, **kwargs)


# === Helper views for development/introspection

def describe_workflow(request, workflow_id):
    w = get_workflow(workflow_id)
    if w is None:
        raise Http404
    return TemplateResponse(request, 'yawf/describe_workflow.html',
            {'workflow': w})



# -*- coding: utf-8 -*-
import copy
import logging
from types import GeneratorType

from django.db import transaction

from yawf.signals import transition_handled
from yawf.utils import select_for_update
from yawf.config import REVISION_ATTR, USE_SELECT_FOR_UPDATE,\
        TRANSACTIONAL_SIDE_EFFECT
from yawf import get_workflow_by_instance
from yawf.exceptions import OldStateInconsistenceError,\
         ConcurrentRevisionUpdate
from yawf.messages.submessage import Submessage
from yawf.messages.transformation import TransformationResult

logger = logging.getLogger(__name__)


def transition(workflow, obj, message, state_transition,
        extra_context=None,
        transactional_side_effect=TRANSACTIONAL_SIDE_EFFECT,
        need_lock_object=USE_SELECT_FOR_UPDATE):
    '''
    Function-dispatcher that allows to control the performing of
    side-effect actions.

    :param transactional_side_effect:
        If `transactional_side_effect` is True, then side effect will be
        performed by :py:func:`transactional_transition`.

        Otherwise, it will be performed after handler commit.

    For other parameters and return values see
    :py:func:`transactional_transition`
    '''

    new_obj, transition_result, effect_result = transactional_transition(
            workflow, obj, message,
            state_transition,
            extra_context=extra_context,
            transactional_side_effect=transactional_side_effect,
            need_lock_object=need_lock_object)

    if not transactional_side_effect:
        effect_result = effect_result()

    return new_obj, transition_result, effect_result


@transaction.commit_on_success
def transactional_transition(workflow, obj, message, state_transition,
        extra_context=None,
        transactional_side_effect=True,
        need_lock_object=True):
    '''
    Performs an extended state transition for object `obj`. Uses
    `commit_on_success` to wrap itself in single transaction.

    :param workflow:
        :py:class:`yawf.base.WorkflowBase` instance, representing object's
        workflow.
    :param obj:
        Workflow aware object.
    :param message:
        :py:class:`yawf.messages.Message` instance.
    :param state_transition:
        callable that takes a single parameter: `obj` and must perform
        all state changes on it.

        Callable might return generator. In such case it will be iterated
        by special routine that filters out instances of
        :py:class:`yawf.messages.submessage.Submessage` class and dispatches
        them as messages within parent transaction context. Any other values
        yielded by generator will be accumulated and returned as a second
        element of result tuple.

        If callable returns something besides generator, the return value will
        be returned as a second element of result tuple.
    :param transactional_side_effect:
        Boolean flag. If `True`, then side effect actions will be performed
        just after state_transition func in single transaction. Otherwise,
        deferred side effect list will be returned (i.e. callable that
        will actually evaluate side effects and return a list of results)
    :return:
        Tuple with three values:
         * A new instance of workflow aware object (possibly changed after a
           state transition)
         * State transition result (list in the case of generator-based
           `state_transition` func, arbitrary object otherwise)
         * Side effect results (either list or a callable to evaluate that list)
    '''

    old_revision = getattr(obj, REVISION_ATTR, None)
    old_state = getattr(obj, workflow.state_attr_name)
    obj_id = obj.id

    # We select for update object because since THIS point we cares
    # about serialization of access to our: we are going to change it's state
    if need_lock_object:
        locked_obj = select_for_update(
                workflow.model_class.objects.filter(id=obj_id)).get()
        locked_revision = getattr(locked_obj, REVISION_ATTR, None)

        # Checking that revision wasn't updated while we worked with object
        # without locking
        if old_revision is not None and locked_revision != old_revision:
            raise ConcurrentRevisionUpdate(workflow.id, obj_id, old_state)

        # Additional checking of state consistency. Matters only if revision
        # check is disabled (getattr above returned None)
        locked_old_state = getattr(locked_obj, workflow.state_attr_name)
        if locked_old_state != old_state:
            raise OldStateInconsistenceError(obj_id,
                    old_state, locked_old_state)
    else:
        locked_obj = copy.copy(obj)

    # All ok, perform db changes as transaction
    transition_result = state_transition(locked_obj)

    new_obj = None

    # If action returned generator, evaluating it using special function
    if isinstance(transition_result, GeneratorType):
        handler_result, pending_calls, new_obj = _iterate_transition_result(
            transition_result, message, locked_obj)
    else:
        handler_result = transition_result
        pending_calls = []

    if new_obj is None:
        new_obj = locked_obj

    # Sending signal
    transition_handled.send(
            sender=workflow.id,
            workflow=workflow,
            message=message,
            instance=obj,
            new_instance=new_obj,
            transition_result=handler_result)

    performed_effects, deferred_effects = perform_side_effect(
                                    obj,
                                    new_obj,
                                    message=message,
                                    workflow=workflow,
                                    handler_result=handler_result,
                                    extra_context=extra_context)


    # decide to evaluate side effect actions now or defer to caller
    if transactional_side_effect:
        side_effect_result = (performed_effects +
                                list(deferred_effects) +
                                map(apply, pending_calls))
    else:
        side_effect_result = lambda: (performed_effects +
                                list(deferred_effects) +
                                map(apply, pending_calls))

    new_state = getattr(new_obj, workflow.state_attr_name)
    logger.info("Performed state transition of object %d: %s -> %s",
            obj_id, old_state, new_state)
    return new_obj, handler_result, side_effect_result


def perform_side_effect(old_obj, new_obj,
        message, workflow=None, extra_context=None, handler_result=None):

    if workflow is None:
        workflow = get_workflow_by_instance(new_obj)

    old_state = getattr(old_obj, workflow.state_attr_name)
    new_state = getattr(new_obj, workflow.state_attr_name)

    (transactional_effects,
            deferrable_effects) = workflow.library\
                                        .get_effects_for_transition(
                                            old_state, new_state, message.id)

    if not transactional_effects and not deferrable_effects:
        logger.info(u"Effect undefined: object id %s, state %s -> %s",
                new_obj.id, old_state, new_state)
        return [], []

    if extra_context is None:
        extra_context = {}

    effect_kwargs = dict(
        old_obj=old_obj,
        obj=new_obj,
        sender=message.sender,
        params=message.params,
        message_spec=message.spec,
        extra_context=extra_context,
        handler_result=handler_result,
    )

    if transactional_effects:
        performed = [
            _perform_side_effect(effect, effect_kwargs)
            for effect in deferrable_effects]
    else:
        performed = []

    if deferrable_effects:
        deferred = (
            _perform_side_effect(effect, effect_kwargs)
            for effect in deferrable_effects)
    else:
        deferred = []
    return performed, deferred


def _perform_side_effect(effect, kwargs):
    effect_result = effect(**kwargs)
    if isinstance(effect_result, GeneratorType):
        return list(effect_result)
    else:
        return effect_result


def _iterate_transition_result(transition_result, message, obj):

    pending_calls = []
    handler_result = []
    new_obj = None

    try:
        yielded_value = transition_result.next()
    except StopIteration:
        return handler_result, pending_calls, new_obj

    while True:
        to_send = None

        if isinstance(yielded_value, Submessage):
            sub_obj, _sub_result, side_effects_performer =\
                                        yielded_value.dispatch(
                                            parent_obj=obj,
                                            parent_message=message)
            pending_calls.append(side_effects_performer)
            to_send = sub_obj
        elif isinstance(yielded_value, TransformationResult):
            new_obj = yielded_value.new_obj
        else:
            handler_result.append(yielded_value)

        try:
            yielded_value = transition_result.send(to_send)
        except StopIteration:
            break

    return handler_result, pending_calls, new_obj



# encoding: utf-8
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models

class Migration(SchemaMigration):

    def forwards(self, orm):
        
        # Adding unique constraint on 'MessageLog', fields ['uuid']
        db.create_unique('message_log_messagelog', ['uuid'])


    def backwards(self, orm):
        
        # Removing unique constraint on 'MessageLog', fields ['uuid']
        db.delete_unique('message_log_messagelog', ['uuid'])


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        'auth.user': {
            'Meta': {'object_name': 'User'},
            'date_joined': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'email': ('django.db.models.fields.EmailField', [], {'max_length': '75', 'blank': 'True'}),
            'first_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'groups': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Group']", 'symmetrical': 'False', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'True'}),
            'is_staff': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'is_superuser': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'last_login': ('django.db.models.fields.DateTimeField', [], {'default': 'datetime.datetime.now'}),
            'last_name': ('django.db.models.fields.CharField', [], {'max_length': '30', 'blank': 'True'}),
            'password': ('django.db.models.fields.CharField', [], {'max_length': '128'}),
            'user_permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'}),
            'username': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '30'})
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        'message_log.messagelog': {
            'Meta': {'ordering': "('created_at',)", 'object_name': 'MessageLog'},
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'related_name': "'message_logs_instance'", 'to': "orm['contenttypes.ContentType']"}),
            'created_at': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'db_index': 'True', 'blank': 'True'}),
            'group_uuid': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '36', 'null': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'initiator_content_type': ('django.db.models.fields.related.ForeignKey', [], {'blank': 'True', 'related_name': "'message_logs_initiator'", 'null': 'True', 'to': "orm['contenttypes.ContentType']"}),
            'initiator_object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True', 'null': 'True', 'blank': 'True'}),
            'message': ('django.db.models.fields.CharField', [], {'max_length': '32'}),
            'message_params': ('django.db.models.fields.TextField', [], {}),
            'object_id': ('django.db.models.fields.PositiveIntegerField', [], {'db_index': 'True'}),
            'parent_uuid': ('django.db.models.fields.CharField', [], {'db_index': 'True', 'max_length': '36', 'null': 'True', 'blank': 'True'}),
            'revision': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['reversion.Revision']", 'null': 'True', 'blank': 'True'}),
            'uuid': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '36', 'db_index': 'True'}),
            'workflow_id': ('django.db.models.fields.CharField', [], {'default': "''", 'max_length': '64', 'db_index': 'True'})
        },
        'reversion.revision': {
            'Meta': {'object_name': 'Revision'},
            'comment': ('django.db.models.fields.TextField', [], {'blank': 'True'}),
            'date_created': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'manager_slug': ('django.db.models.fields.CharField', [], {'default': "'default'", 'max_length': '200', 'db_index': 'True'}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['auth.User']", 'null': 'True', 'blank': 'True'})
        }
    }

    complete_apps = ['message_log']



class RevisionManager(object):

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def bind_revision(self, obj, attrname='revision'):
        pass



# -*- coding: utf-8 -*-
from operator import attrgetter

from yawf import get_workflow_by_instance


def get_allowed_messages(sender, obj, cache=None):
    workflow = get_workflow_by_instance(obj)

    if cache is None:
        cache = dict((c, c(obj, sender)) for c in
                workflow.get_message_checkers_by_state(obj.state))

    for checker, message in workflow.get_available_messages(obj.state):
        if checker(obj, sender, cache=cache):
            yield message


def get_allowed_messages_for_many(sender, objects):

    # TODO: Build a cache for checkers that are independent from object
    #       (use only sender arg). It can be done only if a checker
    #       will be able to mark itself as `sender-only`.

    allowed_map = {}
    for obj in objects:
        allowed_map[obj] = list(get_allowed_messages(sender, obj, cache=None))

    return allowed_map


def get_message_specs_for_many(sender, objects):

    allowed_map = get_allowed_messages_for_many(sender, objects)

    for obj in objects:
        workflow = get_workflow_by_instance(obj)
        allowed_map[obj] = map(workflow.get_message_spec, allowed_map[obj])

    return allowed_map


def get_message_specs(sender, obj):

    workflow = get_workflow_by_instance(obj)
    return [workflow.get_message_spec(mid)
                for mid in get_allowed_messages(sender, obj)]


def is_valid_message(obj, message_id):
    workflow = get_workflow_by_instance(obj)

    return workflow.is_valid_message(message_id, obj.state)


class AllowedWrapper(object):

    def __init__(self, sender, obj, specs=None):
        self.sender = sender
        self.obj = obj
        self._specs = specs
        self._messages_lookup = None

    def _init_lookup(self):
        if self._specs is None:
            self._specs = get_message_specs(self.sender, self.obj)
        self._messages_lookup = dict(
            (message.id, message)
            for message in self._specs
        )

    def __getitem__(self, message_id):
        if self._messages_lookup is None:
            self._init_lookup()

        return self._messages_lookup.get(message_id)

    def __contains__(self, message_id):
        if self._messages_lookup is None:
            self._init_lookup()
        return message_id in self._messages_lookup

    def __iter__(self):
        if self._specs is None:
            self._init_lookup()
        return iter(sorted(self._specs, key=attrgetter('rank')))


def annotate_with_allowed_messages(sender, objects):

    allowed_map = get_message_specs_for_many(sender, objects)
    for obj, specs in allowed_map.iteritems():
        obj.allowed_messages = AllowedWrapper(sender, obj, specs=specs)



from datetime import datetime, date
from operator import attrgetter
from functools import partial

from django.db import models
from django.core.serializers.json import DjangoJSONEncoder
from django.utils.functional import Promise
from django.db.models.fields import FieldDoesNotExist
from django.core import serializers
from django.utils.simplejson import dumps as sj_dumps, loads
from django.utils.translation import force_unicode

from yawf.handlers import SerializibleHandlerResult


class CustomJSONEncoder(DjangoJSONEncoder):

    def default(self, o):
        if isinstance(o, Promise):
            return force_unicode(o)
        elif isinstance(o, datetime):
            return o.strftime("%Y-%m-%dT%H:%M:%S")
        elif isinstance(o, date):
            return o.strftime("%Y-%m-%d")
        elif isinstance(o, models.Model):
            if hasattr(o, 'natural_key'):
                return o.natural_key()
            else:
                return force_unicode(o.pk)
        elif isinstance(o, models.query.QuerySet):
            return map(attrgetter('pk'), o)
        elif isinstance(o, set):
            return list(o)
        elif isinstance(o, SerializibleHandlerResult):
            return o.get_serializible_value()
        else:
            return super(CustomJSONEncoder, self).default(o)

dumps = partial(sj_dumps, cls=CustomJSONEncoder)


def serialize(instance):
    serializer = serializers.get_serializer("json")()
    serializer.serialize([instance], ensure_ascii=False)
    return serializer.getvalue()


def deserialize_to_dict(content):
    item = loads(content)[0]
    return item['fields']


def deserialize(content, format_='json'):
    try:
        deserialized = serializers.deserialize(format_, content).next()
    except FieldDoesNotExist:
        return False, deserialize_to_dict(content)
    else:
        return True, deserialized


def json_converter(attr_name, getter=None, setter=None):
    """
    Returns property object which wraps given attr_name with json load/dump
    """

    if getter is not None:
        def json_getter(instance):
            json_str = getattr(instance, attr_name)
            return getter(loads(json_str)) if json_str is not None else None
    else:
        def json_getter(instance):
            json_str = getattr(instance, attr_name)
            return loads(json_str) if json_str is not None else None

    if setter is not None:
        def json_setter(instance, dict_):
            setattr(instance, attr_name, setter(dumps(dict_)))
    else:
        def json_setter(instance, dict_):
            setattr(instance, attr_name, dumps(dict_))

    return property(json_getter, json_setter)



from django.db import models
from django.contrib.contenttypes import generic


class RevisionModelMixin(models.Model):

    class Meta:
        abstract = True

    _has_revision_support = True

    revision = models.PositiveIntegerField(default=0,
        db_index=True, editable=False)

    versions = generic.GenericRelation('reversion.Version')

    def save(self, *args, **kwargs):
        self.revision += 1
        super(RevisionModelMixin, self).save(*args, **kwargs)



import logging
from collections import Iterable

from django.db import models
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic


from yawf import serialize_utils as json
from yawf.handlers import SerializibleHandlerResult

logger = logging.getLogger(__name__)


class MessageLog(models.Model):

    class Meta:

        get_latest_by = 'created_at'
        ordering = ('created_at',)

    uuid = models.CharField(max_length=36,
        db_index=True, unique=True)

    # Date of log
    created_at = models.DateTimeField(auto_now_add=True,
            db_index=True)

    # Generic foreign key to logged instance
    content_type = models.ForeignKey(ContentType,
            related_name='message_logs_instance')
    object_id = models.PositiveIntegerField(db_index=True)
    instance = generic.GenericForeignKey()

    initiator_content_type = models.ForeignKey(ContentType,
            related_name='message_logs_initiator',
            null=True, blank=True)
    initiator_object_id = models.PositiveIntegerField(
            db_index=True, null=True, blank=True)
    initiator = generic.GenericForeignKey(
            'initiator_content_type', 'initiator_object_id')

    message = models.CharField(max_length=32)
    message_params = models.TextField()
    message_params_dehydrated = models.TextField(default='')

    transition_result = models.TextField(default='')

    workflow_id = models.CharField(max_length=64, db_index=True, default='')
    parent_uuid = models.CharField(max_length=36,
            db_index=True, null=True, blank=True)
    group_uuid = models.CharField(max_length=36,
            db_index=True, null=True, blank=True)

    revision_content_type = models.ForeignKey(ContentType,
            null=True, blank=True, related_name='message_logs_revision')
    revision_id = models.PositiveIntegerField(
            db_index=True, null=True, blank=True)
    revision = generic.GenericForeignKey('revision_content_type', 'revision_id')

    deserialized_params = json.json_converter(
        'message_params')
    deserialized_params_dehydrated = json.json_converter(
        'message_params_dehydrated')
    deserialized_transition_result = json.json_converter(
        'transition_result')

    @staticmethod
    def serialize_params(params):
        return json.dumps(params)


def log_message(sender, **kwargs):
    message = kwargs['message']
    instance = kwargs['new_instance']
    transition_result = kwargs['transition_result']

    initiator = message.sender\
                if isinstance(message.sender, models.Model) else None

    create_dict = dict(
        uuid=message.unique_id,
        message=message.id,
        message_params=MessageLog.serialize_params(message.params),
        workflow_id=sender,
        instance=instance,
        parent_uuid=message.parent_message_id,
        group_uuid=message.message_group,
    )

    if initiator:
        create_dict['initiator'] = initiator

    log_record = MessageLog(**create_dict)
    log_record.deserialized_params = message.params
    log_record.deserialized_params_dehydrated = message.dehydrated_params

    if isinstance(transition_result, Iterable):
        log_record.deserialized_transition_result = filter(
            lambda x: isinstance(x, SerializibleHandlerResult),
            transition_result)
    elif isinstance(transition_result, SerializibleHandlerResult):
        log_record.deserialized_transition_result = [transition_result]

    log_record.save()

    return log_record


def main_record_for_revision(revision):
    ct = ContentType.objects.get_for_model(type(revision))
    return MessageLog.objects.get(
        revision_content_type=ct,
        revision_id=revision.pk,
        parent_uuid__isnull=True)



# -*- coding: utf-8 -*-
from yawf import get_workflow_by_instance


def get_allowed_resources(sender, obj):

    workflow = get_workflow_by_instance(obj)

    check_result = dict((c, c(obj, sender)) for c in
            workflow.get_checkers_by_state(obj.state))

    for resource in workflow.get_available_resources(obj.state):
        if resource.permission_checker(obj, sender, cache=check_result):
            yield resource


def get_resource(obj, resource_id):
    workflow = get_workflow_by_instance(obj)

    return workflow.get_resource(obj.state, resource_id)



# Django settings for yawf_sample project.

DEBUG = True
TEMPLATE_DEBUG = DEBUG

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'test.db',                      # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}

SITE_ID = 1

# Make this unique, and don't share it with anybody.
SECRET_KEY = 'ufq^a%n=9#nbs(_p09c5gvqt(f-7td3$h8tmfbl)(1o9p)226u'

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'yawf_sample.urls'

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'yawf',
    'yawf.message_log',
    'yawf_sample.simple',
    'reversion',
    'django.contrib.admin',
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.CallbackFilter',
            'callback': lambda r: not DEBUG
        }
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
        },
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
        'yawf': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': True,
        }
    }
}

YAWF_CONFIG = {
    'DYNAMIC_WORKFLOW_ENABLED': True,
    'MESSAGE_LOG_ENABLED': True,
    'USE_SELECT_FOR_UPDATE': False,
}

SOUTH_TESTS_MIGRATE = False



from django.db import transaction

from yawf.config import DEFAULT_START_MESSAGE, WORKFLOW_TYPE_ATTR
from yawf import get_workflow, get_workflow_by_instance
from yawf import dispatch
from yawf.exceptions import WorkflowNotLoadedError, CreateValidationError
from yawf.base import WorkflowBase


class CreationAwareWorkflow(WorkflowBase):

    create_form_cls = None
    create_form_template = None

    def __init__(self, *args, **kwargs):
        super(CreationAwareWorkflow, self).__init__(*args, **kwargs)

    def get_create_form_cls(self):

        if self.create_form_cls is None:
            return form_for_model(self.model_class)
        elif isinstance(self.create_form_cls, basestring):
            return class_by_dotted_name(self.create_form_cls)
        else:
            return self.create_form_cls

    def instance_fabric(self, sender, cleaned_data):
        return self.model_class(**cleaned_data)

    def post_create_hook(self, sender, cleaned_data, instance):
        pass


@transaction.commit_on_success
def create(workflow_type, sender, raw_parameters):
    workflow = get_workflow(workflow_type)
    if workflow is None:
        raise WorkflowNotLoadedError(workflow_type)

    create_form_cls = workflow.get_create_form_cls()
    form = create_form_cls(raw_parameters)
    if form.is_valid():

        instance = workflow.instance_fabric(sender, form.cleaned_data)
        # Ensure that we will create, not update
        instance.id = None
        # Set workflow type
        setattr(instance, WORKFLOW_TYPE_ATTR, workflow_type)
        instance.save()

        workflow.post_create_hook(sender, form.cleaned_data, instance)

        return instance
    else:
        raise CreateValidationError(form.errors)


def start_workflow(obj, sender, start_message_params=None):
    if start_message_params is None:
        start_message_params = {}
    workflow = get_workflow_by_instance(obj)

    if isinstance(workflow.start_workflow, basestring):
        start_message_id = workflow.start_workflow
    elif callable(workflow.start_workflow):
        start_message_id  = workflow.start_workflow(obj, sender)
    else:
        start_message_id  = DEFAULT_START_MESSAGE

    return dispatch.dispatch(obj, sender, start_message_id,
                                                start_message_params)


def form_for_model(model_cls):
    from django.forms.models import modelform_factory
    return modelform_factory(model_cls)


class InvalidDottedNameError(ValueError):
    pass


def class_by_dotted_name(dotted_name):
    from django.utils import importlib
    try:
        # Trying to import the given backend, in case it's a dotted path
        mod_path, cls_name = dotted_name.rsplit('.', 1)
        mod = importlib.import_module(mod_path)
        cls = getattr(mod, cls_name)
    except (AttributeError, ImportError, ValueError):
        raise InvalidDottedNameError("Could not find class '%s'" % dotted_name)
    else:
        return cls



from django.conf.urls import patterns, include, url

urlpatterns = patterns('simple.views',
    url(r'^window/(?P<pk>\d+)/resize/$', 'window_resize'),
    url(r'^window/(?P<pk>\d+)/maximize/$', 'window_maximize'),
)



from coded_exceptions import CodedException


class YawfException(CodedException):
    pass


class WorkflowNotLoadedError(YawfException):
    pass


class WorkflowAlreadyRegisteredError(YawfException):
    pass


class OldStateInconsistenceError(YawfException):
    pass


class IllegalStateError(YawfException):
    pass


class UnhandledMessageError(YawfException):

    @property
    def context(self):
        return {'unhandled': self.args[0]}


class WrongHandlerResultError(YawfException):
    pass


class PermissionDeniedError(YawfException):
    pass


class ResourcePermissionDeniedError(PermissionDeniedError):
    pass


class MessageValidationError(YawfException):

    code = 'yawf_validation_error'

    def __init__(self, validator):
        self.validator = validator

    @property
    def context(self):
        return self.validator.errors


class CreateValidationError(MessageValidationError):
    pass


class MessageSpecNotRegisteredError(YawfException):
    pass


class GroupPathEmptyError(YawfException):
    pass


class MessageIgnored(YawfException):

    @property
    def context(self):
        message = self.args[0]
        return {"message_id": message.id,
                "message_params": message.params}


class NoAvailableMessagesError(YawfException):
    pass


class ResourceNotFoundError(YawfException):
    pass


class ConcurrentRevisionUpdate(YawfException):
    pass



# -*- coding: utf-8 -*-
import warnings

from yawf import _register_workflow
from yawf.config import INITIAL_STATE, DEFAULT_START_MESSAGE
from yawf import permissions
from yawf.exceptions import (
    UnhandledMessageError,
)
from yawf.library import Library


def proxy_method(proxy_obj_attr, proxy_attr):

    def proxy_method(self, *args, **kwargs):
        proxy_obj = getattr(self, proxy_obj_attr)
        return getattr(proxy_obj, proxy_attr)(*args, **kwargs)

    return proxy_method


class WorkflowMeta(type):

    def __new__(cls, name, bases, attrs):
        if ('extra_valid_states' in attrs and
                (len(bases) != 1 or (bases[0] is not object))):
            # warn about using extra_valid_states in derived classes
            attrs['states'] = attrs['extra_valid_states']
            attrs.pop('extra_valid_states')
            warnings.warn(
                'Use of extra_valid_states is DEPRECATED. Change your code to use states instead.',
                DeprecationWarning)

        attrs['_library'] = Library(attrs.get('registrants'))
        return super(WorkflowMeta, cls).__new__(cls, name, bases, attrs)


class WorkflowBase(object):
    '''
    Basic class that defines request processing as extended fsm.

    State machine has states, actions, receives messages with parameters,
    know how to validate parameters basing on state and message id. State
    machine also has information about permission checks that sender must
    pass to be able to deliver his message.

    Logic splitted in several submodules of workflow package. This is
    basic class to inherit, that has methods to register callbacks and
    workflow context.
    '''

    __metaclass__ = WorkflowMeta

    rank = 0
    initial_state = INITIAL_STATE
    state_choices = None
    default_permission_checker = permissions.allow_to_all
    verbose_name = None
    verbose_state_names = None
    state_attr_name = 'state'
    registrants = ()

    model_class = None

    exportable_fields = ('rank', 'verbose_name',)
    # message id or callable that returns message context to start workflow
    start_workflow = DEFAULT_START_MESSAGE

    @property
    def extra_valid_states(self):
        warnings.warn(
            'Use of extra_valid_states is DEPRECATED. Change your code to use states instead.',
            DeprecationWarning)
        return self.states

    def __init__(self, inherit_behaviour=False, id=None):

        if self.state_choices:
            self.verbose_state_names = dict(self.state_choices)
        elif self.verbose_state_names:
            self.state_choices = self.verbose_state_names.items()

        self.states = set(self.verbose_state_names.keys())
        self._valid_states = self.states.union([self.initial_state])

        if id is not None:
            self.id = id


        self.inherit_behaviour = inherit_behaviour

        if inherit_behaviour:
            self._library = Library.proxy_library(self._library)

        self._library.default_permission_checker = self.default_permission_checker
        self._library.states = self.states
        self._library.valid_states = self._valid_states

        super(WorkflowBase, self).__init__()
        _register_workflow(self)

    def is_valid_state(self, state):
        return state in self._valid_states

    def is_valid_message(self, state, message_id):
        try:
            self._library.get_handler(state, message_id)
        except UnhandledMessageError:
            return False
        else:
            return True

    @property
    def library(self):
        return self._library

    register_action = register_action_obj = proxy_method('_library', 'effect')
    register_handler = proxy_method('_library', 'handler')
    register_message = proxy_method('_library', 'message')
    register_resource = proxy_method('_library', 'resource')
    register_message_by_form = proxy_method('_library', 'message_by_form')

    get_checkers_by_state = proxy_method('_library',
            'get_checkers_by_state')
    get_message_checkers_by_state = proxy_method('_library',
            'get_message_checkers_by_state')
    get_resource_checkers_by_state = proxy_method('_library',
            'get_resource_checkers_by_state')
    get_available_messages = proxy_method('_library',
            'get_available_messages')

    get_handler = proxy_method('_library', 'get_handler')
    get_action = proxy_method('_library', 'get_effect')
    get_possible_actions = proxy_method('_library', 'get_possible_effects')
    get_possible_effects = proxy_method('_library', 'get_possible_effects')
    get_message_spec = proxy_method('_library', 'get_message_spec')
    get_message_specs = proxy_method('_library', 'get_message_specs')
    get_possible_message_ids = proxy_method('_library', 'get_possible_message_ids')
    get_resource = proxy_method('_library', 'get_resource')
    get_available_resources = proxy_method('_library', 'get_available_resources')
    get_available_messages = proxy_method('_library', 'get_available_messages')

    def validate(self):

        assert self.model_class is not None
        #assert self._deferred or self._handlers
        #assert self._message_specs
        assert self.states



from django.utils.translation import ugettext_lazy as _

from yawf.messages.spec import MessageSpec


def message_spec_fabric(id, verb=None, rank=0, base_spec=MessageSpec, **attrs):
    # NOTE: reflect old behaviour when MessageSpec was registered as class
    # and this function was a class fabric.
    attrs.update({
        'rank': rank,
        'verb': verb,
        'id': id})
    return base_spec(**attrs)


class BasicCancelMessage(MessageSpec):
    id = 'cancel'
    verb = _('cancel')
    rank = 1000


class BasicDeleteMessage(MessageSpec):
    id = 'delete'
    verb = _('delete')
    rank = 1001


class BasicEditMessage(MessageSpec):

    id = 'edit'
    verb = _('change')
    rank = 900

    @staticmethod
    def params_wrapper(params):
        return {'edit_fields': params}


class BasicStartMessage(MessageSpec):
    id = 'start_workflow'
    verb = _('create')



# -*- coding: utf-8 -*-

from django import forms
from django.template import loader as template_loader
from django.template import Context

from yawf import get_workflow, get_workflow_by_instance
from yawf.exceptions import WorkflowNotLoadedError, NoAvailableMessagesError
from yawf.messages.allowed import get_allowed_messages


def get_create_form_html(workflow_type, sender=None):
    workflow = get_workflow(workflow_type)
    if workflow is None:
        raise WorkflowNotLoadedError(workflow_type)

    form = workflow.create_form_cls()
    if not workflow.create_form_template:
        t = template_loader.select_template(
                ('workflows/%s/create_form.html' % workflow_type,
                 'workflows/create_form.html'))
    else:
        t = template_loader.get_template(workflow.create_form_template)

    context = Context({'form': form, 'workflow_type': workflow_type})
    return t.render(context)


def get_object_as_html(obj, sender):
    workflow_type = obj.workflow_type
    allowed_context = get_allowed_messages(sender, obj)

    t = template_loader.select_template(
            ('workflows/%s/object_%s.html' % (workflow_type, obj.state),
                'workflows/%s/object.html' % workflow_type,
                'workflows/object.html'))

    dict_context = {'object': obj}
    dict_context.update(allowed_context)
    context = Context(dict_context)
    return t.render(context)


def get_action_form_html(obj, sender):
    workflow_type = obj.workflow_type
    workflow = get_workflow_by_instance(obj)
    allowed_context = get_allowed_messages(sender, obj)

    t = template_loader.select_template(
            ('workflows/%s/form_%s.html' % (workflow_type, obj.state),
                'workflows/%s/form.html' % workflow_type,
                'workflows/form.html'))

    allowed_messages = allowed_context['allowed_messages']
    if not allowed_messages:
        raise NoAvailableMessagesError(obj.id, sender)

    if callable(workflow.formcls_factory):
        form_cls = workflow.formcls_factory(allowed_messages)
        form = form_cls(instance=obj)
    else:
        # get all form subclasses from message validators, throw away duplicates
        bases = tuple(set(ms.validator_cls for ms in allowed_messages
                            if issubclass(ms.validator_cls, forms.BaseForm)))
        if bases:
            # join them all in one mixin class
            mixin_form_cls = type('WorkflowObjectMixinForm', bases, {})
            # instantiate it and put instance as argument (requires ModelForm subclass to be in bases)
            if issubclass(mixin_form_cls, forms.ModelForm):
                form = mixin_form_cls(instance=obj)
            else:
                form = mixin_form_cls(initial=obj.__dict__)
        else:
            form = forms.Form()

    dict_context = {'form': form, 'instance': obj}
    dict_context.update(allowed_context)
    context = Context(dict_context)
    return t.render(context)



from django.test import TestCase

from yawf.messages.spec import MessageSpec
from yawf.messages.common import message_spec_fabric

__all__ = ('MessageSpecsTestCase',)


class MessageSpecsTestCase(TestCase):

    def test_grouping(self):

        class Spec1(MessageSpec):
            id = 'edit'

        self.assertFalse(Spec1().is_grouped)

        class SpecWithGrouping(MessageSpec):
            id = 'edit__personality__name'
        self.assertTrue(SpecWithGrouping().is_grouped)

    def test_without_id(self):

        class WrongSpec(MessageSpec):
            verb = 'push me!'

        self.assertRaises(ValueError, WrongSpec)

    def test_validator(self):

        class CustomValidator:
            __init__ = lambda self, data: None
            is_valid = lambda self: True
            cleaned_data = {'buz': 'fuz'}

        class Spec(MessageSpec):

            id = 'test'
            validator_cls = CustomValidator

        v = Spec().validator_cls({})
        self.assertTrue(v.is_valid())
        self.assertItemsEqual(
            v.cleaned_data,
            {'buz': 'fuz'})

        class Spec2(Spec):

            class Validator:
                __init__ = lambda self, data: None
                is_valid = lambda self: True
                cleaned_data = {'foo': 'bar'}

        v = Spec2().validator_cls({})
        self.assertTrue(v.is_valid())
        self.assertItemsEqual(
            v.cleaned_data,
            {'foo': 'bar'})

    def test_repr(self):

        class SpecWithoutVerb(MessageSpec):

            id = 'foo'

        self.assertEqual(unicode(SpecWithoutVerb()), 'foo')

        class SpecWithVerb(SpecWithoutVerb):

            verb = 'foobar'

        self.assertEqual(unicode(SpecWithVerb()), 'foobar')

    def test_fabric(self):

        spec = message_spec_fabric(id='cancel')
        self.assertEqual(spec.id, 'cancel')
        self.assertTrue(isinstance(spec, MessageSpec))



from django.test import TestCase
from django.utils.unittest import skipIf
import reversion

import yawf
from yawf.exceptions import MessageSpecNotRegisteredError, UnhandledMessageError
import yawf.creation
import yawf.dispatch
from yawf.handlers import Handler
from yawf.revision.utils import (
    diff_fields, versions_diff, deserialize_revision, previous_version)
from yawf.message_log.models import main_record_for_revision
from yawf.messages.spec import MessageSpec
from yawf.allowed import get_allowed

yawf.autodiscover()
from .models import Window, WINDOW_OPEN_STATUS


class WorkflowTestMixin(object):

    workflow_id = None

    def get_workflow(self):
        return yawf.get_workflow(self.workflow_id)

    def test_workflow_registered(self):
        workflow = self.get_workflow()
        self.assertIsNotNone(workflow)

    def test_initial_handler(self):
        w = self.get_workflow()
        if isinstance(w.start_workflow, basestring):
            start_handler = w.get_handler(w.initial_state, w.start_workflow)
            self._test_handler_obj(start_handler)

    def test_initial_message(self):
        w = self.get_workflow()
        if isinstance(w.start_workflow, basestring):
            spec = w.get_message_spec(w.start_workflow)
            self._test_message_spec(spec)

    def _test_handler_obj(self, handler):
        self.assertIsInstance(handler, yawf.handlers.Handler)

    def _test_message_spec(self, spec):
        self.assertIsInstance(spec, yawf.messages.spec.MessageSpec)
        self.assertIsInstance(spec.id, basestring)
        self.assertTrue(hasattr(spec, 'validator_cls'))
        self.assertTrue(hasattr(spec.validator_cls, 'is_valid'))

    def test_validate(self):
        w = self.get_workflow()
        w.validate()


class SimpleWorkflowTest(WorkflowTestMixin, TestCase):

    workflow_id = 'simple'
    sender = '__sender__'

    def test_creation(self):

        self.assertRaises(
            yawf.exceptions.CreateValidationError,
            lambda: yawf.creation.create(self.workflow_id, self.sender, {}))

        window = yawf.creation.create(
            self.workflow_id, self.sender,
            {
                'title': 'Main window',
                'width': 500,
                'height': 300,
            })
        new_instance, handler_effect, side_effect = yawf.creation.start_workflow(window, self.sender)
        self.assertEqual(window.id, new_instance.id)
        self.assertFalse(window is new_instance)

    def test_grouped_action(self):
        window, _, _ = self._new_window(width=500, height=300)
        self.assertEqual(window.width, 500)
        self.assertEqual(window.height, 300)

        resized_window, handler_effects, effects = yawf.dispatch.dispatch(window, self.sender,
            'edit__resize', dict(width=200, height=400))
        self.assertEqual(resized_window.width, 200)
        self.assertEqual(resized_window.height, 400)

        resized_window = Window.objects.get(id=window.id)
        self.assertEqual(resized_window.width, 200)
        self.assertEqual(resized_window.height, 400)

        self.assertListEqual(effects, ['edit_effect', 'resize_effect'])

    def test_multimessage(self):
        window, _, _ = self._new_window(width=500, height=300)
        child1, _, _ = self._new_window(parent=window)
        child2, _, _ = self._new_window(parent=window)

        window, handler_effects, _ = yawf.dispatch.dispatch(
                            window, self.sender, 'minimize_all')

        self.assertTrue(isinstance(handler_effects, list))
        self.assertEqual(len(handler_effects), 3)
        window, child1, child2 = handler_effects
        self.assertEqual(window.open_status, WINDOW_OPEN_STATUS.MINIMIZED)
        self.assertEqual(child1.open_status, WINDOW_OPEN_STATUS.MINIMIZED)
        self.assertEqual(child2.open_status, WINDOW_OPEN_STATUS.MINIMIZED)

    def test_revision_deserialization(self):
        window, _, _ = self._new_window(width=500, height=300)
        self.assertEqual(window.revision, 2)

        child1, _, _ = self._new_window(parent=window)
        child2, _, _ = self._new_window(parent=window)

        window, _, _ = yawf.dispatch.dispatch(
                            window, self.sender, 'minimize_all')

        versions = reversion.get_for_object(window)
        self.assertEqual(len(versions), 2)
        last_version = versions[0]
        message_revision = last_version.revision

        # checking log record
        log_record = main_record_for_revision(message_revision)
        self.assertEqual(log_record.message, 'minimize_all')
        self.assertEqual(log_record.object_id, window.id)

        # checking revision
        rev = deserialize_revision(message_revision)
        versions = rev.get_versions_for_record(log_record)
        self.assertEqual(len(versions), 1)
        version = versions[0]
        self.assertEqual(last_version, version)
        previous = previous_version(version)
        diff = versions_diff(previous, version, full=True)
        self.assertItemsEqual(diff[0],
            {
                'field_name': 'open_status',
                'old': 'normal',
                'new': 'minimized',
                'field_verbose_name': 'open status'
            })
        self.assertItemsEqual(diff[1],
            {
                'field_name': 'revision',
                'old': '2',
                'new': '3',
                'field_verbose_name': 'revision'
            })

    def test_revision_diff(self):
        window, _, _ = self._new_window(width=500, height=300)
        self.assertEqual(window.revision, 2)

        resized_window, _, _ = yawf.dispatch.dispatch(window, self.sender,
            'edit__resize', dict(width=200, height=300))
        self.assertEqual(resized_window.revision, 3)

        versions = reversion.get_for_object(window)
        self.assertEqual(len(versions), 2)
        new_rev = versions[0]
        old_rev = versions[1]

        diff = list(diff_fields(old_rev, new_rev))
        self.assertItemsEqual(diff, ['width'])

        diff = versions_diff(old_rev, new_rev)
        self.assertItemsEqual(diff,
            [
                {
                    'field_name': 'width',
                    'old': 500,
                    'new': 200,
                    'field_verbose_name': 'width'
                },
            ])

    def test_allowed(self):
        window, _, _ = self._new_window()
        allowed = get_allowed(self.sender, window)
        self.assertItemsEqual(allowed.keys(), ['allowed_messages', 'allowed_resources'])


    def test_view_handling(self):
        window, _, _ = self._new_window(width=500, height=300)
        self.assertEqual(window.width, 500)
        self.assertEqual(window.height, 300)

        self.client.post(
            '/simple/window/%d/resize/' % window.id,
            {
                'width': 200,
                'height': 400,
            })

        window = Window.objects.get(pk=window.id)
        self.assertEqual(window.width, 200)
        self.assertEqual(window.height, 400)

        self.assertEqual(window.open_status, 'normal')
        r = self.client.post('/simple/window/%d/maximize/' % window.id)
        self.assertEqual(r.status_code, 200)

        window = Window.objects.get(pk=window.id)
        self.assertEqual(window.open_status, 'maximized')


    def _new_window(self, title='Main window', width=500, height=300,
            parent=None):
        window = yawf.creation.create(
            self.workflow_id, self.sender,
            {
                'title': title,
                'width': width,
                'height': height,
                'parent': parent.id if parent is not None else None,
            })
        return yawf.creation.start_workflow(window, self.sender)


def which(name):
    """Searches for name in exec path and returns full path (from pygraphviz)"""
    import os
    import glob
    paths = os.environ["PATH"]
    if os.name == "nt":
        exe = ".exe"
    else:
        exe = ""
    for path in paths.split(os.pathsep):
        match = glob.glob(os.path.join(path, name + exe))
        if match:
            return match[0]
    return None


class BuiltinViewTest(TestCase):

    def test_describe(self):
        response = self.client.get('/describe/simple/')
        self.assertEqual(response.status_code, 200)

        response = self.client.get('/describe/some_nonexist_workflow/')
        self.assertEqual(response.status_code, 404)

    @skipIf(which('dot') is None,
            "graphviz is not installed")
    def test_handlers_graph(self):
        response = self.client.get('/describe/simple/graph/handlers/')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.has_header('Content-Type'))
        self.assertEqual(response['Content-Type'], 'image/png')
        self.assertTrue(len(response.content) > 1024)

    @skipIf(which('dot') is None,
            "graphviz is not installed")
    def test_effects_graph(self):
        response = self.client.get('/describe/simple/graph/effects/')
        self.assertEqual(response.status_code, 200)
        self.assertTrue(response.has_header('Content-Type'))
        self.assertEqual(response['Content-Type'], 'image/png')
        self.assertTrue(len(response.content) > 1024)


class MessageGroupsTest(WorkflowTestMixin, TestCase):

    workflow_id = 'message_groups'
    sender = '__sender__'

    def test_grouped_handlers_registration(self):
        wf = self.get_workflow()
        spec = wf.get_message_spec('hover')
        self.assertIsInstance(spec, MessageSpec)

        with self.assertRaises(MessageSpecNotRegisteredError):
            wf.get_message_spec('hover__subhover')

        spec = wf.get_message_spec('edit')
        self.assertIsInstance(spec, MessageSpec)

        spec = wf.get_message_spec('edit__title')
        self.assertIsInstance(spec, MessageSpec)

        handler = wf.get_handler(WINDOW_OPEN_STATUS.NORMAL, 'hover')
        self.assertIsInstance(handler, Handler)

        with self.assertRaises(UnhandledMessageError):
            wf.get_handler(WINDOW_OPEN_STATUS.NORMAL, 'hover__subhover')

        handler = wf.get_handler(WINDOW_OPEN_STATUS.NORMAL, 'edit__title')
        self.assertIsInstance(handler, Handler)

        handler = wf.get_handler(WINDOW_OPEN_STATUS.NORMAL, 'edit')
        self.assertIsInstance(handler, Handler)


class MinimalWorkflowTest(WorkflowTestMixin, TestCase):

    workflow_id = 'minimal'
    sender = '__sender__'



from django.core.management.base import BaseCommand

from yawf import get_workflow
from yawf.graph import build_effects_graph


class Command(BaseCommand):

    def handle(self, workflow_id, **options):
        w = get_workflow(workflow_id)
        print build_effects_graph(w).to_string()



# -*- coding: utf-8 -*-


class BasePermissionChecker(object):

    def __init__(self, *checkers):
        self._checkers = list(checkers)
        super(BasePermissionChecker, self).__init__()

    def fill_cache(self, obj, sender):
        cache = {}
        for c in set(self.get_atomical_checkers()):
            cache[c] = c(obj, sender)
        return cache

    def get_atomical_checkers(self):
        for c in self._checkers:
            if isinstance(c, BasePermissionChecker):
                if c is self:
                    # recursive checker TODO: log with warning
                    continue
                for child_c in c.get_atomical_checkers():
                    yield child_c
            else:
                 yield c

    def perform_child_checker(self, checker, obj, sender, cache):
        if isinstance(checker, BasePermissionChecker):
            return checker(obj, sender, cache=cache)
        else:
            cache_result = cache.get(checker, None)
            return (cache_result
                if cache_result is not None
                else checker(obj, sender))

    def add_checker(self, checker):
        self._checkers.append(checker)

    def __and__(self, other):
        return AndChecker(self, other)

    __rand__ = __and__

    def __invert__(self):
        return NotChecker(self)

    def __or__(self, other):
        return OrChecker(self, other)

    __ror__ = __or__


class AndChecker(BasePermissionChecker):

    def __call__(self, obj, sender, cache=None):
        if cache is None:
            cache = self.fill_cache(obj, sender)

        return all(
            self.perform_child_checker(c, obj, sender, cache=cache)
            for c in self._checkers)

    def __iand__(self, other):
        self.add_checker(other)
        return self


class OrChecker(BasePermissionChecker):

    def __call__(self, obj, sender, cache=None):
        if cache is None:
            cache = self.fill_cache(obj, sender)

        return any(
            self.perform_child_checker(c, obj, sender, cache=cache)
            for c in self._checkers)

    def __ior__(self, other):
        self.add_checker(other)
        return self


class NotChecker(BasePermissionChecker):

    def __init__(self, checker):
        self._invertable_checker = checker
        super(NotChecker, self).__init__(checker)

    def __call__(self, obj, sender, cache=None):
        if cache is None:
            cache = self.fill_cache(obj, sender)

        return not self.perform_child_checker(self._invertable_checker,
                obj, sender, cache=cache)

    def __invert__(self):
        return self._invertable_checker


# shortcut
C = AndChecker

# basic checkers
allow_to_all = OrChecker(lambda obj, sender: True)
restrict_to_all = OrChecker(lambda obj, sender: False)



class EmptyValidator(object):
    '''
    Class with django-form behaviour that filters out all incoming data,
    always becomes valid and returns empty dict as cleaned data.
    '''

    def __init__(self, *args, **kwargs):
        super(EmptyValidator, self).__init__()

    def is_valid(self):
        return True

    @property
    def cleaned_data(self):
        return {}


class MessageSpec(object):
    '''
    Specifies a message that can be passed to workflow.

    Basic class for one of main entities.
    Message spec knows:
      * it's unique ``id'';
      * how to represent spec for humans (using ``verb'' in __unicode__);
      * how to validate parameters for this message (using ``validator_cls'');
      * priority (``rank'' attribute).
    '''
    # unique (in scope of single workflow) message id (typically str)
    id = None
    # human-friendly name for action
    verb = None
    # validator class -- with django.forms.Form interface
    validator_cls = EmptyValidator
    # rank used to sort specs
    rank = 0

    id_grouper = '__'
    is_grouped = False

    def __init__(self, **attrs):
        super(MessageSpec, self).__init__()

        if hasattr(self, 'Validator'):
            self.validator_cls = self.Validator

        for attr, value in attrs.iteritems():
            setattr(self, attr, value)

        grouper = self.id_grouper
        message_id = self.id

        if message_id is None:
            raise ValueError(
                'You must specify not-None id for MessageSpec subclass')
        else:
            if grouper in message_id:
                self.group_path = message_id.split(grouper)
                self.is_grouped = True
            else:
                self.group_path = [message_id]
                self.is_grouped = False

            if self.verb is None:
                self.verb = self.id

    def params_wrapper(self, params):
        '''
        Wrapper used to wrap cleaned_data returned by validator_cls
        before passing it to handler.

        With wrapper one has additional layer to transform structure of
        cleaned_data dict to make it more convenient to use.

        For example, see yawf.messages.common.BasicEditMessage class.
        '''
        return params

    def dehydrate_params(self, obj, message):
        '''
        Method, that returns dehydrated message params for serialization.
        '''
        return None

    def __unicode__(self):
        return unicode(self.verb)

    def __repr__(self):
        return u'<%s: %s>' % (self.__class__.__name__, self)



from django import forms

from yawf.creation import CreationAwareWorkflow
from yawf.messages.common import message_spec_fabric, BasicStartMessage, MessageSpec
from yawf.messages.submessage import Submessage, RecursiveSubmessage

from yawf.actions import SideEffect
from yawf.handlers import SimpleStateTransition, Handler, EditHandler, LoopHandler, StartWorkflowHandlerBase
from yawf.utils import make_common_updater
from yawf.annotation import annotate_handler

from yawf_sample.simple.models import Window, WINDOW_OPEN_STATUS

class MessageGroupsTestWorkflow(CreationAwareWorkflow):

    id = 'message_groups'
    verbose_name = 'Workflow to test message grouping'
    model_class = Window
    state_choices = WINDOW_OPEN_STATUS.choices
    state_attr_name = 'open_status'


workflow = MessageGroupsTestWorkflow()
register = workflow.library

register.message(BasicStartMessage)

@register.handler
class StartWorkflowHandler(StartWorkflowHandlerBase):
    state_to = WINDOW_OPEN_STATUS.NORMAL


class WindowEditHandler(EditHandler):

    message_group = 'edit'

register.handler(WindowEditHandler)


@register.message
class EditMessageSpec(MessageSpec):

    id = 'edit'

    class Validator(forms.ModelForm):
        class Meta:
            model = Window
            fields = ('title', 'width', 'height')


@register.message
class EditTitleMessageSpec(MessageSpec):

    id = 'edit__title'

    class Validator(forms.ModelForm):
        class Meta:
            model = Window
            fields = ('title',)


@register.message
class EditSizeMessageSpec(MessageSpec):

    id = 'edit__resize'

    class Validator(forms.ModelForm):
        class Meta:
            model = Window
            fields = ('width', 'height')


@register.message
class HoverMessageSpec(MessageSpec):

    id = 'hover'

    class Validator(forms.Form):

        pos_x = forms.IntegerField()
        pos_y = forms.IntegerField()


@register.handler
class HoverHandler(LoopHandler):

    message_id = 'hover'
    message_group = 'hover'



from django.conf.urls import patterns, include, url

# Uncomment the next two lines to enable the admin:
from django.contrib import admin
admin.autodiscover()

import yawf
yawf.autodiscover()

urlpatterns = patterns('',
    url(r'^simple/', include('simple.urls')),
    url(r'^describe/(?P<workflow_id>\w+)/$', 'yawf.views.describe_workflow',
        name='describe'),
    url(r'^describe/(?P<workflow_id>\w+)/graph/handlers/$',
        'yawf.graph_views.handlers_graph'),
    url(r'^describe/(?P<workflow_id>\w+)/graph/effects/$',
        'yawf.graph_views.effects_graph',
        ),

    # Uncomment the admin/doc line below to enable admin documentation:
    # url(r'^admin/doc/', include('django.contrib.admindocs.urls')),

    # Uncomment the next line to enable the admin:
    url(r'^admin/', include(admin.site.urls)),
)



import collections
import inspect
from functools import wraps
from operator import attrgetter

from django.utils.datastructures import MergeDict
from django.utils.importlib import import_module

from yawf.effects import SideEffect
from yawf.handlers import Handler
from yawf.resources import WorkflowResource
from yawf.exceptions import (
    UnhandledMessageError,
    IllegalStateError,
    MessageSpecNotRegisteredError,
    GroupPathEmptyError,
)
from yawf.messages.spec import MessageSpec
from yawf.messages.common import message_spec_fabric
from yawf.permissions import OrChecker
from yawf.utils import metadefaultdict, maybe_list


def merge_container(container_name, container_fabric, parent_container):

    # XXX: need to test before extensive use

    if inspect.isclass(container_fabric) and issubclass(container_fabric, dict):
        basic_container = container_fabric()
        if container_name == '_message_specs':
            class NewMergeDict(MergeDict):
                __setitem__ = basic_container.__setitem__

            container = NewMergeDict(basic_container, parent_container)
        else:
            container = parent_container

        return container
    else:
        return parent_container


def touches_index(method):

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self._is_index_built:
            if not self._is_imported_registrants:
                self.import_registrants()
            self.rebuild_index()
        return method(self, *args, **kwargs)

    return wrapper


def need_imports(method):

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self._is_imported_registrants:
            self.import_registrants()
        return method(self, *args, **kwargs)

    return wrapper


class Library(object):

    _is_index_built = False
    _is_imported_registrants = False

    _containers = (
        ('_resources', dict),
        ('_resources_by_state', dict),
        ('_message_specs', dict),
        ('_message_groups', dict),
        ('_handler_patterns', list),
        ('_effect_patterns', list),
        ('_resource_checkers_index', metadefaultdict(set)),
        ('_registered_message_id_set', set),
    )

    _index_containers = (
        ('_handler_index', metadefaultdict(list)),
        ('_handler_message_index', metadefaultdict(metadefaultdict(list))),
        ('_handler_state_index', metadefaultdict(metadefaultdict(list))),
        ('_message_checkers_index', metadefaultdict(set)),
        ('_effect_index', metadefaultdict(list)),
        ('_deferrable_effect_index', metadefaultdict(list)),
        ('_transactional_effect_index', metadefaultdict(list)),
        ('_possible_effect_index', metadefaultdict(list)),
    )

    def __init__(self, registrants=()):
        self._init_containers()
        if registrants:
            self._registrants = registrants
        else:
            self._registrants = ()
        super(Library, self).__init__()

    def import_registrants(self):
        for dotted_name in self._registrants:
            import_module(dotted_name)
        self._is_imported_registrants = True

    @classmethod
    def proxy_library(cls, base, *bases):

        self = cls()
        for container_name, container_fabric in self._containers:

            merged_container = merge_container(
                container_name,
                container_fabric,
                getattr(base, container_name))

            for extra_base in bases:
                merged_container = merge_container(
                    container_name,
                    lambda: merged_container,
                    getattr(extra_base, container_name))

            setattr(self, container_name, merged_container)

        return self

    def message(self, message_spec):

        to_return = message_spec

        if inspect.isclass(message_spec) and\
                issubclass(message_spec, MessageSpec):
            message_spec = message_spec()

        if not isinstance(message_spec, MessageSpec):
            raise TypeError('Message spec must be an instance or a subclass of MessageSpec')

        # register in flat registry
        if message_spec.id in self._registered_message_id_set:
            raise ValueError("Message spec already registered for message '%s'" %
                    (message_spec.id,))
        self._registered_message_id_set.add(message_spec.id)
        self._message_specs[message_spec.id] = message_spec

        # handle message grouping
        if message_spec.is_grouped:
            group_path, final_id = message_spec.group_path[:-1],\
                                   message_spec.group_path[-1]
            group_dict = self._message_groups

            for group in group_path:
                group_dict = group_dict.setdefault(
                    group, {})

            if final_id in group_dict:
                raise ValueError("Message spec already registered for message '%s' in group '%s'" %
                    (final_id, group_path))

            group_dict[final_id] = message_spec

        if self._is_index_built:
            self.rebuild_index()

        return to_return

    def message_by_form(self, message_id=None, message_id_list=None,
            base_spec=MessageSpec, **attrs):

        if message_id is None:
            if message_id_list is None:
                raise ValueError("You must specify either message_id or message_id_list")
        else:
            message_id_list = [message_id]

        def registrator(message_validator):

            for message_id in message_id_list:
                spec_cls = message_spec_fabric(
                    message_id,
                    base_spec=base_spec,
                    validator_cls=message_validator,
                    **attrs)

                self.message(spec_cls)
            return message_validator

        return registrator

    def resource(self, resource_id=None, description=None,
            available_in_states=None,
            permission_checker=None,
            slug=None):

        if permission_checker is None:
            permission_checker = OrChecker(self.default_permission_checker)
        elif not isinstance(permission_checker, collections.Iterable):
            permission_checker = OrChecker(permission_checker)
        else:
            permission_checker = OrChecker(*permission_checker)

        if available_in_states is None:
            available_in_states = self.states

        if not available_in_states:
            raise ValueError("available_in_states cannot be empty")

        def registrator(handler):

            if resource_id in self._resources:
                raise ValueError("Resource with that name already registered")

            resource = WorkflowResource(handler, resource_id=resource_id,
                    description=description,
                    permission_checker=permission_checker,
                    slug=slug)

            self._resources[resource_id] = resource

            for state in available_in_states:
                # add handler to lookup table by (state, message_id)
                resources = self._resources_by_state.setdefault(state, {})
                resources[resource_id] = resource
                self._resource_checkers_index[state].update(
                    permission_checker.get_atomical_checkers())

            return handler

        return registrator

    def handler(self, message_id=None, **options):
        '''
        Returns decorator to register handler for message_id when fsm in
        one of the states in states_from and sender passed permission_checker.

        If states_from is None we register given handler for every valid state.

        permission_checker may be either callable or iterable of callables.
        In second case, all checkers are joined in single lambda function
        with logical OR rule.

        Permission checkers from iterable are indexed by state so that one
        can get all distinct checkers that must be evaluated to know what
        messages single sender can pass to fsm (see get_checkers_by_state
        method and functions in yawf.messages.allowed).
        '''
        return self._meta_register(Handler, self._register_handler_obj,
                message_id=message_id, **options)

    def effect(self,
            message_id=None, **options):

        return self._meta_register(SideEffect, self._register_effect_obj,
                message_id=message_id, **options)

    def rebuild_index(self):
        self._init_index_containers()

        # Build index for handlers
        for pattern in self._handler_patterns:
            message_id_list, group_path, states_from, handler = pattern

            message_id_list = list(message_id_list)
            if group_path:
                message_id_list.extend(
                    self._get_message_ids_by_path(group_path))

            if not states_from:
                states_from = self.states

            for message_id in message_id_list:
                for state in states_from:

                    self._handler_index[(state, message_id)].append(handler)
                    self._handler_message_index[message_id][state].append(handler)
                    self._handler_state_index[state][message_id].append(handler)
                    # add checker to checkers index
                    self._message_checkers_index[state].update(
                        handler.permission_checker.get_atomical_checkers())

        # Build index for effects
        for pattern in self._effect_patterns:
            message_id_list, group_path, states_to, states_from, effect = pattern

            message_id_list = list(message_id_list)
            if group_path:
                message_id_list.extend(
                    self._get_message_ids_by_path(group_path))

            if not states_from:
                states_from = self.states

            if not states_to:
                states_to = self.states

            # if message specs not specified, stick to any registered message
            if not message_id_list:
                message_id_list = self._registered_message_id_set

            for message_id in message_id_list:

                for state_from in states_from:

                    key = (state_from, message_id)
                    self._possible_effect_index[key].append(effect)

                    for state_to in states_to:
                        key = (state_from, state_to, message_id)
                        self._effect_index[key].append(effect)
                        if effect.is_transactional:
                            self._transactional_effect_index[key].append(effect)
                        else:
                            self._deferrable_effect_index[key].append(effect)

        self._is_index_built = True

    def get_handler(self, state, message_id):
        return self.get_handlers(state, message_id)[0]

    @touches_index
    def get_handlers(self, state, message_id):
        '''
        Return Handler instances to handle ``message_id'' for workflow in state
        ``state''.

        :param state:
            Current basic state of workflow.
        :param message_id:
            id of incoming message.

        If there are no handler for this message_id, raises
        UnhandledMessageError(message_id).
        '''
        key = (state, message_id)
        handlers = self._handler_index.get(key)

        if not handlers:
            if state is None:
                return self._handler_message_index.get(message_id, [])
            elif message_id is None:
                return self._handler_state_index.get(state, [])
            else:
                raise UnhandledMessageError(message_id)

        return handlers

    def get_effect(self, from_state, to_state, message_id):
        effects = self.get_effects(from_state, to_state, message_id)
        if effects:
            return effects[0]

    @touches_index
    def get_effects(self, from_state, to_state, message_id):
        key = (from_state, to_state, message_id)
        return self._effect_index.get(key)

    @touches_index
    def get_effects_for_transition(self, from_state, to_state, message_id):
        key = (from_state, to_state, message_id)
        return (self._transactional_effect_index.get(key),
                self._deferrable_effect_index.get(key))

    @touches_index
    def get_possible_effects(self, from_state, message_id):
        return self._possible_effect_index.get((from_state, message_id)) or []

    @need_imports
    def get_message_spec(self, message_id):
        '''
        Get registered message spec class with given id.

        Id can be iterable representing group path to the desired message.
        '''
        if not isinstance(message_id, basestring) and\
                isinstance(message_id, collections.Iterable):
            return self._get_message_spec_by_path(message_id)

        message_spec = self._message_specs.get(message_id)
        if message_spec is None:
            raise MessageSpecNotRegisteredError(message_id)
        return message_spec

    @need_imports
    def get_message_specs(self):
        return self._message_specs

    @need_imports
    def get_possible_message_ids(self):
        return self._registered_message_id_set

    def get_resource(self, state, resource_id):
        '''
        Return resource object or None if there is no resource with given id
        for given state.
        '''
        lookup_result = self._resources_by_state.get(state)
        if lookup_result is None:
            if state not in self.valid_states:
                raise IllegalStateError(state)
            return None

        return lookup_result.get(resource_id)

    def get_available_resources(self, state):
        lookup_result = self._resources_by_state.get(state)
        if lookup_result is None:
            if state not in self.valid_states:
                raise IllegalStateError(state)
            return ()

        return lookup_result.itervalues()

    def get_available_messages(self, state):

        lookup_result = self.get_handlers(state, None)

        if not lookup_result:
            if state not in self.valid_states:
                raise IllegalStateError(state)
            return ()

        return ((_handler.permission_checker, message_id)
                for message_id, _handlers in
                    lookup_result.iteritems()
                        for _handler in _handlers
                        if message_id in self._registered_message_id_set)

    def get_checkers_by_state(self, state):
        return self.get_message_checkers_by_state(state)\
                .union(self.get_resource_checkers_by_state(state))

    @touches_index
    def get_message_checkers_by_state(self, state):
        return self._message_checkers_index[state]

    def get_resource_checkers_by_state(self, state):
        return self._resource_checkers_index[state]

    @touches_index
    def iter_handlers(self):
        return self._handler_index.iteritems()

    @touches_index
    def iter_effects(self):
        return self._effect_index.iteritems()

    def _meta_register(self, reg_cls, registrator, message_id, **options):

        if inspect.isclass(message_id) and\
                issubclass(message_id, reg_cls):
            registrator(message_id())
            return message_id

        if isinstance(message_id, reg_cls):
            registrator(message_id)
            return message_id

        def _registrator(func):

            if message_id is None:
                _message_id = func.__name__
            else:
                _message_id = message_id

            obj = reg_cls(
                message_id=_message_id,
                **options)

            obj.set_performer(func)
            registrator(obj)
            return func

        if inspect.isfunction(message_id):
            func = message_id
            message_id = None
            return _registrator(func)
        else:
            return _registrator

    def _register_handler_obj(self, handler):

        if handler.permission_checker is None:
            handler.permission_checker = OrChecker(
                self.default_permission_checker)

        group_path = handler.message_group
        message_id_list = maybe_list(handler.message_id)

        self._handler_patterns.append(
            (message_id_list, group_path, handler.states_from, handler)
        )

        if self._is_index_built:
            self.rebuild_index()

        return handler

    def _register_effect_obj(self, effect):

        group_path = effect.message_group
        message_id_list = maybe_list(effect.message_id)
        states_to, states_from = effect.states_to, effect.states_from

        self._effect_patterns.append(
            (message_id_list, group_path, states_to, states_from, effect)
        )

        if self._is_index_built:
            self.rebuild_index()

        return effect

    def _get_message_spec_by_path(self, group_path):
        cur_level = self._message_groups
        for group_id in group_path:
            cur_level = cur_level.get(group_id, None)
            if not cur_level:
                raise GroupPathEmptyError(group_path)
        return cur_level

    def _get_message_ids_by_path(self, group_path):
        cur_level = self._message_groups
        for group_id in group_path:
            cur_level = cur_level.get(group_id, None)
            if not cur_level:
                return []
        return map(attrgetter('id'), cur_level.itervalues())

    def _init_containers(self):
        for container_name, container_fabric in self._containers:
            setattr(self, container_name, container_fabric())

    def _init_index_containers(self):
        for container_name, container_fabric in self._index_containers:
            setattr(self, container_name, container_fabric())



from yawf.base import WorkflowBase
from yawf.handlers import StartWorkflowHandlerBase
from yawf.messages.common import BasicStartMessage

from yawf_sample.simple.models import Window, WINDOW_OPEN_STATUS


class MinimalWorkflow(WorkflowBase):

    id = 'minimal'
    state_choices = WINDOW_OPEN_STATUS.choices
    model_class = Window

workflow = MinimalWorkflow()
register = workflow.library

register.message(BasicStartMessage)

class StartWorkflowHandler(StartWorkflowHandlerBase):
    state_to = WINDOW_OPEN_STATUS.NORMAL

register.handler(StartWorkflowHandler)



import itertools as itr

from django.contrib.contenttypes.models import ContentType
from reversion.models import Version

from yawf.utils import model_diff_fields, model_diff
from yawf.serialize_utils import deserialize


def diff_fields(old, new):
    # If content type doesn't match, we can't tell the difference
    if new.content_type_id != old.content_type_id:
        return None

    is_new_full, new = deserialize(new.serialized_data)
    is_old_full, old = deserialize(old.serialized_data)

    if is_new_full and is_old_full:
        return model_diff_fields(old.object, new.object)


def versions_diff(old, new, full=False):
    '''
    Calculate (as possible) difference between two revisions.

    :return:
        List of dicts with keys:
            * ``field_name'';
            * ``field_verbose_name'';
            * ``old'';
            * ``new''.
    '''
    # If content type doesn't match, we can't tell the difference
    if old.content_type_id != new.content_type_id:
        return None
    is_new_full, new = deserialize(new.serialized_data)
    is_old_full, old = deserialize(old.serialized_data)

    if is_new_full and is_old_full:
        return model_diff(old.object, new.object, full=full)


def previous_version(version):
    try:
        return (Version.objects
                    .filter(
                        content_type__id=version.content_type_id,
                        object_id_int=version.object_id_int,
                        pk__lt=version.pk)
                    .order_by('-pk')[0])
    except IndexError:
        return None


class DeserializedRevision(object):

    def __init__(self, revision):
        super(DeserializedRevision, self).__init__()

        self.revision = revision

        versions = (revision.version_set
                            .select_related('content_type')
                            .order_by('content_type'))

        grouped = itr.groupby(
            versions,
            lambda v: '.'.join(
                [v.content_type.app_label, v.content_type.model]))

        self._index = dict(
            (grouper, dict(
                (o.object_id_int, o)# SimpleLazyObject(lambda: o.object_version))
                for o in group))
            for grouper, group in grouped)

    def get_version_for_object(self, content_type, object_id):
        dotted_model_name = '%s.%s' % (
            content_type.app_label, content_type.model)
        return self._index.get(dotted_model_name, {}).get(object_id)

    def get_versions_for_record(self, record):
        model_cls = record.content_type.model_class()
        version = self.get_version_for_object(
            record.content_type, record.object_id)

        versions = [version]

        if model_cls._meta.parents:
            for parent_model in model_cls._meta.parents.keys():
                v = self.get_version_for_object(
                    ContentType.objects.get_for_model(parent_model),
                    record.object_id)
                if v is not None:
                    versions.append(v)

        return versions

    def get_versions_for_content_type(self, ct):
        dotted_model_name = '.'.join(
            [ct.app_label,
             ct.model])
        return self._index.get(dotted_model_name, {}).values()

    def get_object_for_record(self, record):
        return deserialize_inherited_versions(self.get_versions_for_record(record))


def deserialize_version(version):
    return deserialize(version.serialized_data, version.format)


def fill_child_with_parent_fields(child, parent):
    for field in parent._meta.fields:
        setattr(child, field.name, getattr(parent, field.name))


def deserialize_inherited_versions(versions):
    objs = [obj
        for (deserialization_result, obj)
        in map(deserialize_version, versions)
        if deserialization_result
    ]
    child = objs[0].object
    for obj in objs[1:]:
        fill_child_with_parent_fields(child, obj.object)
    return child

deserialize_revision = DeserializedRevision



import os

from setuptools import setup, find_packages, findall

PACKAGE_NAME = 'yawf'
CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
REQUIREMENTS_FILEPATH = os.path.join(CURRENT_DIR, 'requirements', 'default.txt')
README_FILEPATH = os.path.join(CURRENT_DIR, 'README.rst')
VERSION_FILEPATH = os.path.join(CURRENT_DIR, PACKAGE_NAME, 'version.py')


def get_version():
    # populate namespace with __version__
    execfile(VERSION_FILEPATH)
    return locals()['__version__']


def get_requirements():
    with open(REQUIREMENTS_FILEPATH) as fp:
        return fp.read().splitlines()


def get_data_files():
    data_files = filter(
        lambda name: not name.endswith('.py') and not name.endswith('.pyc'),
        findall(PACKAGE_NAME))
    return [x.split(os.sep, 1)[-1] for x in data_files]


def get_long_description():
    return open(README_FILEPATH).read()


setup(
    name = PACKAGE_NAME,
    version = get_version(),
    packages = find_packages(CURRENT_DIR, exclude=('yawf_sample', 'yawf_sample.*')),
    package_data = {'': get_data_files()},

    # Metadata
    author = 'Nikolay Zakharov',
    author_email = 'nikolay@desh.su',
    url = 'https://github.com/freevoid/yawf',
    description = 'Yet Another Workflow Framework',
    long_description = get_long_description(),
    keywords = 'workflow state transition fsm django',
    install_requires = get_requirements(),
    extras_require = {
        'reversion': ['django-reversion>=1.5'],
    },
    license = 'MIT',
    classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Framework :: Django',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
    ],
)



import pygraphviz


def build_effects_graph(workflow):

    g = pygraphviz.AGraph(directed=True, strict=False)

    for state in workflow._valid_states:
        g.add_node(state)

    for (state_from, state_to, message), _effects in workflow.library.iter_effects():
        g.add_edge(state_from, state_to, message)
        e = g.get_edge(state_from, state_to, message)
        e.attr['label'] = message

    return g

build_actions_graph = build_effects_graph


def build_handlers_graph(workflow):

    g = pygraphviz.AGraph(directed=True, strict=False)

    for state in workflow._valid_states:
        g.add_node(state)

    unknown_color = 'gray'
    g.add_node('?')
    unknown_state = g.get_node('?')
    unknown_state.attr['color'] = unknown_color

    for ((state_from, message), handlers) in workflow.library.iter_handlers():
        for handler in handlers:
            if hasattr(handler, 'is_annotated'):
                if len(handler.states_to) > 1:
                    style = 'dashed'
                else:
                    style = 'solid'

                for state_to in handler.states_to:

                    if state_to == '_':
                        state_to = state_from

                    g.add_edge(state_from, state_to, message)
                    e = g.get_edge(state_from, state_to, message)
                    e.attr['label'] = message
                    e.attr['style'] = style
            else:
                g.add_edge(state_from, unknown_state, message)
                e = g.get_edge(state_from, unknown_state, message)
                e.attr['label'] = message
                e.attr['color'] = unknown_color

    return g



import logging
import collections

from yawf.permissions import BasePermissionChecker, OrChecker
from yawf.config import INITIAL_STATE

logger = logging.getLogger(__name__)

__all__ = ['Handler', 'SimpleStateTransition', 'ComplexStateTransition',
        'EditHandler', 'SerializibleHandlerResult']


class Handler(object):
    '''
    Basic class to express state transitions, both simple and extended.

    Handler object is a callable. It gets a workflow object, message sender
    and arbitrary keyword arguments (passed with message) and must return
    new state id based on this information (it can return None and message
    will be ignored).

    If handler returns a callable, then this callable applies as transaction
    on object (to change its extended state). The result of the latter
    callable represents a transition result. If callable returns generator,
    transition routine will iterate over it, collecting the yielded values.
    For more information about state transition, see
    :py:mod:`yawf.state_transition` module.

    Transitioning callable can actually return anything besides generator.
    That is totally an application designer business.
    '''
    message_id = None
    message_group = None
    states_from = None
    permission_checker = None
    defer = True
    replace_if_exists = False
    copy_before_call = False

    def __init__(self, message_id=None, states_from=None,
            message_group=None,
            permission_checker=None):

        if message_id is not None:
            self.message_id = message_id
        if message_group is not None:
            self.message_group = message_group
        if states_from is not None:
            self.states_from = states_from
        if permission_checker is not None:
            self.permission_checker = permission_checker

        message_group = self.message_group

        if message_group is not None:
            if isinstance(message_group, basestring):
                self.message_group = [message_group]
                if self.message_id is None:
                    self.message_id = message_group
            else:
                assert isinstance(message_group, collections.Iterable)
        elif self.message_id is None:
            raise ValueError("message_id must be specified for handler")

        if self.permission_checker is not None:
            self._align_permission_checker()

        super(Handler, self).__init__()

    def _align_permission_checker(self):
        if not isinstance(self.permission_checker, collections.Iterable):
            if not isinstance(self.permission_checker, BasePermissionChecker):
                self.permission_checker = OrChecker(self.permission_checker,)
        else:
            self.permission_checker = OrChecker(*self.permission_checker)

        assert callable(self.permission_checker)

    def __call__(self, obj, sender, **kwargs):
        return self.perform(obj, sender, **kwargs)

    def perform(self, obj, sender, **kwargs):
        logger.warning("Message ignored (by default)")
        return None

    def set_performer(self, handle_func):
        self.perform = lambda obj, sender, **kwargs:\
            handle_func(obj, sender, **kwargs)


class SimpleStateTransition(Handler):

    state_to = None

    def __init__(self, *args, **kwargs):
        self.states_to = [self.state_to]
        self.is_annotated = True
        super(SimpleStateTransition, self).__init__(*args, **kwargs)

    def perform(self, obj, sender, **kwargs):
        return self.state_to


class ComplexStateTransition(Handler):

    def transition(self, obj, sender, **kwargs):
        raise NotImplementedError

    def perform(self, obj, sender, **kwargs):
        return lambda obj: self.transition(obj, sender, **kwargs)


class EditHandler(ComplexStateTransition):

    field_names = None

    def post_hook(self, obj):
        return

    def transition(self, obj, sender, **kwargs):

        # to ensure that we will update, not insert new
        kwargs.pop('id', None)
        kwargs.pop('pk', None)

        field_names = self.field_names
        if field_names:
            for field_name in field_names:
                value = kwargs.get(field_name, '__missing')
                if value != '__missing':
                    setattr(obj, field_name, value)
        else:
            for field_name in (f.name for f in obj._meta.fields):
                value = kwargs.get(field_name, '__missing')
                if value != '__missing':
                    setattr(obj, field_name, value)

            for field_name in (f.name for f in obj._meta.many_to_many):
                value = kwargs.get(field_name)
                if value is not None:
                    setattr(obj, field_name, value)

        obj.save()

        return self.post_hook(obj)


class LoopHandler(Handler):

    def perform(self, obj, sender, **kwargs):
        return getattr(obj, obj.workflow.state_attr_name)


class StartWorkflowHandlerBase(SimpleStateTransition):

    message_id = 'start_workflow'
    states_from = [INITIAL_STATE]


class SerializibleHandlerResult(object):

    type = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        super(SerializibleHandlerResult, self).__init__()

    def get_serializible_value(self):
        return {
            'type': self.type,
            'result': self.kwargs,
        }



from django.contrib import admin
from yawf.message_log.models import MessageLog


class MessageLogAdmin(admin.ModelAdmin):

    ordering = ('-created_at',)
    date_hierarchy = 'created_at'
    list_filter = ('workflow_id', 'message')
    list_display = (
        'id', 'created_at', 'workflow_id', 'object_id', 'message', 'initiator',
        'uuid', 'group_uuid')

admin.site.register(MessageLog, MessageLogAdmin,)



from django import forms

from yawf.creation import CreationAwareWorkflow
from yawf.messages.common import message_spec_fabric, BasicStartMessage, MessageSpec
from yawf.messages.submessage import Submessage, RecursiveSubmessage

from yawf.actions import SideEffect
from yawf.handlers import SimpleStateTransition, Handler, ComplexStateTransition
from yawf.utils import make_common_updater
from yawf.annotation import annotate_handler

from yawf_sample.simple.models import Window, WINDOW_OPEN_STATUS


class SimpleWorkflow(CreationAwareWorkflow):

    id = 'simple'
    verbose_name = 'Just a simple workflow'
    model_class = Window
    state_choices = WINDOW_OPEN_STATUS.choices
    state_attr_name = 'open_status'
    registrants = (
        'yawf_sample.simple.views',
    )

simple_workflow = SimpleWorkflow()


@simple_workflow.register_message_by_form('click')
class ClickForm(forms.Form):

    pos_x = forms.IntegerField()
    pos_y = forms.IntegerField()


@simple_workflow.register_message
class ResizeMessage(MessageSpec):

    id = 'edit__resize'
    verb = 'Resize window'

    class Validator(forms.Form):

        width = forms.IntegerField(min_value=1)
        height = forms.IntegerField(min_value=1)

    @staticmethod
    def params_wrapper(params):
        return {'edit_fields': params}


simple_workflow.register_message(MessageSpec(id='minimize', verb='Minimize window'))
simple_workflow.register_message(MessageSpec(id='maximize'))
simple_workflow.register_message(message_spec_fabric(id='minimize_all'))
simple_workflow.register_message(BasicStartMessage)

@simple_workflow.register_handler
class Start(SimpleStateTransition):

    message_id = 'start_workflow'
    states_from = ['init']
    state_to = 'normal'

@simple_workflow.register_handler
class ToMinimized(SimpleStateTransition):

    message_id = 'minimize'
    states_from = ['normal', 'maximized']
    state_to = 'minimized'

#NOTE: example of complex state transition
@simple_workflow.register_handler
class MinimizeAll(ComplexStateTransition):

    message_id = 'minimize_all'
    states_from = ['normal', 'maximized']

    def transition(self, obj, sender):
        yield (yield RecursiveSubmessage('minimize', sender))
        for child_window in obj.children.all():
            yield (yield Submessage(child_window, 'minimize', sender))

@annotate_handler(states_to=('normal',))
@simple_workflow.register_handler(states_from=['maximized', 'minimized'])
def to_normal(obj, sender):
    return 'normal'

@simple_workflow.register_handler
class Edit(Handler):

    message_group = 'edit'
    states_to = ['_']
    is_annotated = True

    def perform(self, obj, sender, edit_fields):
        return make_common_updater(edit_fields)


@simple_workflow.register_action
class SignalizeEdit(SideEffect):

    message_group = 'edit'

    def perform(self, **kwargs):
        return 'edit_effect'


@simple_workflow.register_action
class SignalizeResize(SideEffect):

    message_id = 'edit__resize'

    def perform(self, **kwargs):
        return 'resize_effect'

#simple_workflow._clean_deferred_chain()



import uuid

from yawf.exceptions import MessageValidationError


class Message(object):

    def __init__(self, sender, message_id,
            raw_params=None, message_group=None, parent_message_id=None,
            clean_params=None):
        self.sender = sender
        self.id = message_id
        self._unique_id = None
        self.raw_params = raw_params if raw_params is not None else {}
        self.clean_params = clean_params if clean_params is not None else None
        self.params = None
        self.dehydrated_params = None
        self.parent_message_id = parent_message_id
        self.message_group = message_group\
            if message_group is not None else self.unique_id
        super(Message, self).__init__()

    @property
    def unique_id(self):
        '''
        Property that returns the unique identifier for message.

        Internally it uses UUID.
        '''
        if not self._unique_id:
            self._unique_id = uuid.uuid1()
        return self._unique_id

    def clean(self, workflow, obj):
        message_spec = workflow.get_message_spec(self.id)

        if self.clean_params is None:
            validator_cls = message_spec.validator_cls
            validator = validator_cls(self.raw_params)

            if validator.is_valid():
                self.clean_params = validator.cleaned_data
            else:
                raise MessageValidationError(validator)

        self.params = message_spec.params_wrapper(self.clean_params)
        # fix id in case it was a list (group path)
        self.id = message_spec.id
        self.spec = message_spec

        return self

    def dehydrate_params(self, workflow, obj):
        if self.params is None:
            raise RuntimeError(
                'Method dehydrate_params cannot be invoked before clean')

        self.dehydrated_params = self.spec.dehydrate_params(obj, self)
        return self


from .submessage import Submessage
from .transformation import TransformationResult
